"""Helpers for computing metrics during optimization.

TODO:
    - Active features should not take into account inactive neurons.
"""
import timeit
from typing import Any, Dict, List, Optional, Tuple

import scnn.private.loss_functions as loss_fns
import scnn.private.models.solution_mappings as sm
import lab
from scnn.private.models import AL_MLP, ConvexMLP
from scnn.private.models.model import Model


def as_list(x: Any) -> List[Any]:
    """Wrap argument into a list if it is not iterable.

    :param x: a (potential) singleton to wrap in a list.
    :returns: [x] if x is not iterable and x if it is.
    """
    # don't treat strings as iterables.
    if isinstance(x, str):
        return [x]

    try:
        _ = iter(x)
        return x
    except TypeError:
        return [x]


def format_recent_metrics(
    metric_dict: Dict[str, Any],
    metrics: Tuple[List[str], List[str], List[str]],
):
    """Print the most recent metric evaluation contained in the metric log.

    :param metric_dict: a dictionary used to record metric information during optimization. This dictionary should
        be initialize by calling 'init_metrics' before being passed to this function.
    :param metrics: a tuple of the form (train_metrics, test_metrics, additional_metrics)
        specifying the metrics to be computed on the training set, test set, and data-independent
        metrics.
    """
    train_metrics, test_metrics, additional_metrics = metrics
    metric_str = ""

    for metric_name in train_metrics:
        metric_str = (
            metric_str
            + f"Train Set {metric_name}: {metric_dict['train_' + metric_name][-1]}, "
        )

    for metric_name in test_metrics:
        metric_str = (
            metric_str
            + f"Test Set {metric_name}: {metric_dict['test_' + metric_name][-1]}, "
        )

    for metric_name in additional_metrics:
        if metric_name == "subproblem_metrics":
            continue

        metric_str = (
            metric_str + f"{metric_name}: {metric_dict[metric_name][-1]}, "
        )

    return metric_str


def merge_metric_lists(
    lm: Tuple[List[str], List[str], List[str]],
    rm: Optional[Tuple[List[str], List[str], List[str]]] = None,
) -> Tuple[List[str], List[str], List[str]]:

    if rm is None:
        return lm

    return (
        list(set(lm[0] + rm[0])),
        list(set(lm[1] + rm[1])),
        list(set(lm[2] + rm[2])),
    )


def init_metrics(
    metric_dict: Dict[str, List[Any]],
    metrics: Tuple[List[str], List[str], List[str]],
):
    """Initialize a metric dictionary with the necessary lists. This should be
    called before using 'update_metrics'.

    :param metric_dict: a (possibly empty) dictionary used to record metric information during optimization.
    :param metrics: a tuple of the form (train_metrics, test_metrics, additional_metrics)
        specifying the metrics to be computed on the training set, test set, and data-independent
        metrics.
    :returns: metric_dict with an empty list at each training/test metric key.
    """

    train_metrics, test_metrics, additional_metrics = metrics

    for metric_name in train_metrics:
        if "train_" + metric_name not in metric_dict:
            metric_dict["train_" + metric_name] = []

    # test metrics
    for metric_name in test_metrics:
        if "test_" + metric_name not in metric_dict:
            metric_dict["test_" + metric_name] = []

    for metric_name in additional_metrics:
        if metric_name not in metric_dict:
            metric_dict[metric_name] = []

    return metric_dict


def update_metrics(
    metric_dict: Dict[str, List[Any]],
    model: Model,
    sp_exit_state: Dict[str, Any],
    train_data: Tuple[lab.Tensor, lab.Tensor],
    test_data: Tuple[lab.Tensor, lab.Tensor],
    metrics: Tuple[List[str], List[str], List[str]],
    train_objective: Optional[lab.Tensor] = None,
    train_grad: Optional[lab.Tensor] = None,
    batch_size: Optional[int] = None,
) -> Dict[str, List[Any]]:
    """Update a dictionary of metrics given a model, a training set, a test
    set, and a list of metrics to use.

    :param metric_dict: a dictionary used to record metric information during optimization. This dictionary should
        be initialize by calling 'init_metrics' before being passed to this function.
    :param model: the model for which to compute the metrics.
    :param sp_exit_state: the exit status of any sub-processes called by the optimizer.
    :param train_data: an (X,y) tuple containing the training dataset.
    :param test_data: an (X_test, y_test) tuple containing the test/validation set.
    :param metrics: a tuple of the form (train_metrics, test_metrics, additional_metrics)
        specifying the metrics to be computed on the training set, test set, and data-independent
        metrics.
    :param batch_size: the batch size to use when computing the objective.
        Defaults to `None' which indicates full-batch.
    :returns: metric_dict with an empty list at each training/test metric key.
    """

    train_metrics, test_metrics, additional_metrics = metrics

    # training metrics
    for metric_name in train_metrics:
        metric_dict = compute_metric(
            metric_name,
            metric_dict,
            "train_" + metric_name,
            model,
            sp_exit_state,
            train_data,
            train_objective,
            train_grad,
            batch_size=batch_size,
        )

    # test metrics
    model.eval()
    for metric_name in test_metrics:
        metric_dict = compute_metric(
            metric_name,
            metric_dict,
            "test_" + metric_name,
            model,
            sp_exit_state,
            test_data,
            batch_size=batch_size,
        )
    model.train()

    for metric_name in additional_metrics:
        metric_dict = compute_metric(
            metric_name,
            metric_dict,
            metric_name,
            model,
            sp_exit_state,
            train_data,
            batch_size=batch_size,
        )

    return metric_dict


def compute_metric(
    metric_name: str,
    metric_dict: Dict[str, Any],
    dict_key: str,
    model: Model,
    sp_exit_state: Dict[str, Any],
    data: Tuple[lab.Tensor, lab.Tensor],
    objective: Optional[lab.Tensor] = None,
    grad: Optional[lab.Tensor] = None,
    batch_size: Optional[int] = None,
) -> Any:
    """Lookup a metric by name and compute it for a given model and dataset.

    :param metric_name: the name of the metric to compute.
    :param metric_dict: a dictionary used to record metric information during optimization. This dictionary should
        be initialize by calling 'init_metrics' before being passed to this function.
    :param dict_key: the key at which the metric should be stored in 'metric_dict'.
    :param model: the model for which the metric should be computed.
    :param sp_exit_state: the exit status of any sub-processes called by the optimizer.
    :param data: an (X,y) tuple containing a training/validation/test set on which to compute the metric.
    :param batch_size: the batch size to use when computing the objective.
        Defaults to `None' which indicates full-batch.
    :returns: the value of the desired metric.
    """
    X, y = data
    metric = None

    if metric_name == "objective":
        if objective is None:
            objective = model.objective(X, y, batch_size=batch_size)
        metric = lab.to_scalar(objective)
    elif metric_name == "base_objective":
        objective = model.objective(
            X, y, batch_size=batch_size, ignore_lagrange_penalty=True
        )
        metric = lab.to_scalar(objective)
    elif metric_name == "squared_error":
        metric = lab.to_scalar(
            loss_fns.squared_error(model(X, batch_size=batch_size), y)
            / y.shape[0]
        )
    elif metric_name == "grad_norm":
        if grad is None:
            grad = model.grad(X, y, batch_size=batch_size)
        metric = lab.to_scalar(lab.sum(grad ** 2))
    elif metric_name == "binary_accuracy":
        metric = lab.to_scalar(
            lab.sum(lab.sign(model(X, batch_size=batch_size)) == y)
            / y.shape[0]
        )
    elif metric_name == "binned_accuracy":
        metric = lab.to_scalar(
            loss_fns.binned_accuracy(model(X, batch_size=batch_size), y, 10)
        )
    elif metric_name == "accuracy":
        if y.shape[1] > 1:
            metric = lab.to_scalar(
                loss_fns.accuracy(model(X, batch_size=batch_size), y)
            )
        else:
            metric = lab.to_scalar(
                lab.sum(lab.sign(model(X, batch_size=batch_size)) == y)
                / y.shape[0]
            )
    elif metric_name == "time_stamp":
        metric = timeit.default_timer()
    elif metric_name == "total_neurons":
        metric = model.p
    elif metric_name == "feature_sparsity":
        reduced_weights = model.weights
        feature_weights = lab.sum(
            lab.abs(reduced_weights),
            axis=tuple(range(len(reduced_weights.shape) - 1)),
        )
        # feature_weights = feature_weights + lab.sum(lab.abs(model.U), axis=-1)

        metric = lab.to_scalar(
            lab.sum(feature_weights == 0.0) / reduced_weights.shape[-1]
        )
    elif metric_name == "active_features":
        reduced_weights = model.weights
        feature_weights = lab.sum(
            lab.abs(reduced_weights),
            axis=tuple(range(len(reduced_weights.shape) - 1)),
        )
        # feature_weights = feature_weights + lab.sum(lab.abs(model.U), axis=-1)
        metric = lab.to_scalar(lab.sum(feature_weights != 0.0))
    elif metric_name == "group_sparsity":
        metric = lab.to_scalar(
            lab.sum(lab.sum(model.weights, axis=-1) == 0.0)
            / (lab.size(model.weights) / model.weights.shape[-1])
        )
    elif metric_name == "active_neurons":
        metric = lab.to_scalar(lab.sum(lab.sum(model.weights, axis=-1) != 0.0))
    elif metric_name == "sparsity":
        metric = lab.to_scalar(
            lab.sum(model.weights == 0) / lab.size(model.weights)
        )
    elif metric_name == "group_norms":
        metric = lab.to_scalar(
            lab.sum(lab.sqrt(lab.sum(model.weights ** 2, axis=-1)))
        )
    elif metric_name == "dual_param_norm":
        try:
            metric = lab.to_scalar(lab.sum(model.i_multipliers ** 2))
        except:
            metric = lab.tensor([0.0])
    elif metric_name == "constraint_gaps":
        try:
            e_gap, i_gap = model.constraint_gaps(X)
        except:
            e_gap = i_gap = lab.tensor([0.0])

        metric = lab.to_scalar(lab.sum(e_gap ** 2 + lab.smax(i_gap, 0) ** 2))
    elif metric_name == "lagrangian_grad":
        try:
            metric = lab.to_scalar(lab.sum(model.lagrangian_grad(X, y) ** 2))
        except:
            metric = 0.0
    elif metric_name == "num_backtracks":
        metric = sp_exit_state.get("attempts", 1) - 1
    elif metric_name == "sp_success":
        metric = sp_exit_state.get("success", True)
    elif metric_name == "step_size":
        metric = lab.to_scalar(sp_exit_state.get("step_size", 0.0))
    elif metric_name == "subproblem_metrics":
        for key, value in sp_exit_state.get("subproblem_metrics", {}).items():
            old_value = metric_dict.get(f"subproblem_{key}", [])
            metric_dict[f"subproblem_{key}"] = old_value + as_list(value)
    elif "nc_" in metric_name:
        nc_model = model

        # compute the non-convex model if one exists
        if isinstance(model, ConvexMLP):
            nc_model = sm.get_nc_formulation(model, remove_sparse=True)

        compute_metric(
            metric_name.split("nc_")[1],
            metric_dict,
            dict_key,
            nc_model,
            sp_exit_state,
            data,
            None,
            None,
            batch_size,
        )

    else:
        raise ValueError(
            f"Metric {metric_name} not recognized! Please register it in 'metrics.py'"
        )
    if metric is not None:
        metric_dict.get(dict_key, []).append(metric)

    return metric_dict



"""End-to-end tests for :module:`scnn.optimize`.

These tests exercise the internal models and optimization routines in :module:`scnn.private` as well as the
interface between the public and private modules :module`scnn.private.interface`.

Most tests are end-to-end and don't check specific termination conditions. Rather, they ensure that the entire 
library works properly when invoked with particular inputs.


TODO:
    -
"""

import unittest
import torch
import numpy as np

from scnn.private.utils.data.synthetic import gen_regression_data

from scnn.optimize import optimize, optimize_model, optimize_path
from scnn.regularizers import NeuronGL1
from scnn.models import ConvexGatedReLU, ConvexReLU
from scnn.solvers import RFISTA, AL
from scnn.metrics import Metrics
from scnn.activations import sample_gate_vectors


class TestDataFormats(unittest.TestCase):
    """Test inputting different data formats different floating-point precisions.

    We support three data formats at the moment:

    1. Python lists and lists of lists.
    2. NumPy arrays
    3. PyTorch tensors located on the CPU device.
    """

    def setUp(self):
        # Generate realizable synthetic classification problem (ie. Figure 1)
        n_train = 100
        n_test = 100
        d = 25
        c = 2
        kappa = 1
        self.max_neurons = 100

        (
            (self.X_train, self.y_train),
            (self.X_test, self.y_test),
            _,
        ) = gen_regression_data(123, n_train, n_test, d, c, kappa=kappa)

        self.regularizer = NeuronGL1(0.0001)

    def test_scalar_targets(self):
        """Test :func:`scnn.optimize.optimize` with scalar targets."""

        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train[:, 0],
            self.X_test,
            self.y_test[:, 0],
            self.regularizer,
        )

    def test_numpy(self):
        """Test :func:`scnn.optimize.optimize` with data as NumPy ndarrays."""

        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

    def test_lists(self):
        """Test :func:`scnn.optimize.optimize` with data in list of lists format."""

        X_train = self.X_train.tolist()
        X_test = self.X_test.tolist()

        y_train = self.y_train.tolist()
        y_test = self.y_test.tolist()

        # train model
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            X_train,
            y_train,
            X_test,
            y_test,
            self.regularizer,
        )

    def test_torch(self):
        """Test :func:`scnn.optimize.optimize` with data as PyTorch tensors."""
        X_train = torch.tensor(self.X_train.tolist())
        X_test = torch.tensor(self.X_test.tolist())

        y_train = torch.tensor(self.y_train.tolist())
        y_test = torch.tensor(self.y_test.tolist())

        # train model
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            X_train,
            y_train,
            X_test,
            y_test,
            self.regularizer,
        )

    def test_no_test_data(self):
        """Test :func:`scnn.optimize.optimize` without any test data."""
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            regularizer=self.regularizer,
        )


class TestFunctionalInterface(unittest.TestCase):
    """Test simple functional interface to `scnn`."""

    def setUp(self):
        # Generate realizable synthetic classification problem (ie. Figure 1)
        n_train = 100
        n_test = 100
        d = 25
        c = 2
        kappa = 1
        self.max_neurons = 100

        (
            (self.X_train, self.y_train),
            (self.X_test, self.y_test),
            _,
        ) = gen_regression_data(123, n_train, n_test, d, c, kappa=kappa)

        self.regularizer = NeuronGL1(0.0001)

    def test_solving_gated_relu(self):
        """Test solving the Gated ReLU problem with neuron-wise GL1 regularization using
        the default solver, R-FISTA.
        """
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

    def test_solving_relu(self):
        """Test solving the ReLU problem with neuron-wise GL1 regularization using
        the default solver, AL method with R-FISTA as a sub-routine.
        """
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )


class TestOOInterface(unittest.TestCase):
    """Test object-oriented interface to `scnn`."""

    def setUp(self):

        n_train = 100
        n_test = 100
        self.d = 5
        self.c = 2
        kappa = 1
        self.max_neurons = 100

        (
            (self.X_train, self.y_train),
            (self.X_test, self.y_test),
            _,
        ) = gen_regression_data(
            123, n_train, n_test, self.d, self.c, kappa=kappa
        )
        self.metrics = Metrics(
            metric_freq=25,
            model_loss=True,
            train_accuracy=True,
            train_mse=True,
            test_mse=True,
            test_accuracy=True,
            neuron_sparsity=True,
        )
        self.lam = 0.0001
        self.regularizer = NeuronGL1(self.lam)
        self.G = sample_gate_vectors(
            np.random.default_rng(123), self.d, self.max_neurons
        )

        # regularization paths
        self.lam_path = np.logspace(-5, -2, 3)
        self.path = [NeuronGL1(lam) for lam in self.lam_path]

    def test_solving_gated_relu(self):
        """Test solving the Gated ReLU problem with neuron-wise GL1 regularization using
        the default solver, R-FISTA.
        """

        # Instantiate convex model and other options.
        model = ConvexGatedReLU(self.G, c=self.c)
        solver = RFISTA(model, tol=1e-6)

        oo_model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

        # check against functional interface
        f_model, _ = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
            seed=123,  # ensures the same activation patterns are used.
        )

        self.assertTrue(
            np.allclose(
                oo_model.get_parameters()[0], f_model.get_parameters()[0]
            )
        )

    def test_solving_gated_relu_path(self):
        """Test solving a regularization path for the Gated ReLU problem."""

        # Instantiate convex model and other options.
        model = ConvexGatedReLU(self.G, c=self.c)
        solver = RFISTA(model, tol=1e-6)

        model_path, _ = optimize_path(
            model,
            solver,
            self.path,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            warm_start=True,
            save_path="test/test_data/gated",
        )

    def test_solving_relu(self):
        """Test solving the ReLU problem with neuron-wise GL1 regularization using
        the default solver, AL method with R-FISTA as a sub-routine.
        """

        model = ConvexReLU(self.G, c=self.c)
        solver = AL(model)

        oo_model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

        # check against functional interface
        f_model, _ = optimize(
            "relu",
            2 * self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
            seed=123,  # ensures the same activation patterns are used.
        )

        self.assertTrue(
            np.allclose(
                oo_model.get_parameters()[0], f_model.get_parameters()[0]
            )
        )

        self.assertTrue(
            np.allclose(
                oo_model.get_parameters()[1], f_model.get_parameters()[1]
            )
        )

    def test_solving_relu_path(self):
        """Test solving a regularization path for the ReLU problem."""

        # Instantiate convex model and other options.
        model = ConvexReLU(self.G, c=self.c)
        solver = AL(model)

        model_path, _ = optimize_path(
            model,
            solver,
            self.path,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            warm_start=True,
            save_path="test/test_data/relu",
        )


if __name__ == "__main__":
    unittest.main()



"""Wrapper for PyTorch Sequential module ensuring that it meets the interface
specific in model.py."""

from typing import Optional, Union, Tuple

import torch

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer
import scnn.private.loss_functions as loss_fns


class SequentialWrapper(torch.nn.Sequential, Model):
    """Model-style wrapper for PyTorch Sequential module that provides an
    interface for objective and gradient computation."""

    activation_history: Optional[torch.Tensor] = None
    weight_history: Optional[torch.Tensor] = None

    def __init__(
        self,
        *modules,
        loss_fn=loss_fns.squared_error,
        regularizer: Optional[Regularizer] = None,
    ):

        super().__init__(*modules)

        self.loss_fn = loss_fn
        self.regularizer = regularizer
        self.layers = modules

    @property
    def weights(self):
        return torch.cat([param.ravel() for param in self.parameters()])

    def forward(self, X: torch.Tensor, batch_size=None):
        return super().forward(X)

    def objective(
        self,
        X: torch.Tensor,
        y: torch.Tensor,
        w: Optional[torch.Tensor] = None,
        ignore_regularizer: bool = False,
        scaling: Optional[float] = None,
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Compute objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: (NOT USED) specific parameter at which to compute the objective.
        :param ignore_regularizer: (optional) whether or not to ignore the regularizer and return
            *only* the un-regularized objective.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :param batch_size: (NOT USED) the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: objective L(f, (X,y)).
        """
        # hard-code squared loss with weight decay

        obj = self.loss_fn(self.__call__(X), y) / self._scaling(y, scaling)

        for layer in self.layers:
            inner_reg = layer.get_regularizer()

            if inner_reg is not None and not ignore_regularizer:
                obj += inner_reg.penalty(layer.parameters())
            elif self.regularizer is not None and not ignore_regularizer:
                obj += self.regularizer.penalty(layer.parameters())

        return obj

    def grad(
        self,
        X: torch.Tensor,
        y: torch.Tensor,
        w: Optional[torch.Tensor] = None,
        ignore_regularizer: bool = False,
        return_model_grad: bool = False,
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> Union[lab.Tensor, Tuple[lab.Tensor, lab.Tensor]]:
        """Compute the gradient of the objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: (NOT USED) specific parameter at which to compute the gradient.
        :param ignore_regularizer: (optional) whether or not to ignore the regularizer and return
            *only* the gradient of the un-regularized objective.
        :param batch_size: (NOT USED) the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: the gradient
        """
        with torch.set_grad_enabled(True):
            # clear current gradients
            self.zero_grad()

            # compute loss and call autodiff engine
            loss = self.objective(X, y, w, ignore_regularizer=ignore_regularizer)
            loss.backward()

        return torch.cat([param.grad.ravel() for param in self.parameters()])

    def sign_patterns(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> Tuple[torch.Tensor, lab.Tensor]:
        """Compute the gradient of the l2 objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: (NOT USED) specific parameter at which to compute the sign patterns.
        :returns: the set of sign patterns active at w or the current models parameters if w is None.
        """
        with torch.no_grad():
            first_layer = next(next(next(self.children()).children()).children())

            # The first layer *must* be a linear map.
            assert isinstance(first_layer, torch.nn.Linear)

            return (
                torch.sign(torch.max(first_layer(X), torch.tensor(0))),
                first_layer.weight,
            )



"""Proximal operators. This module provides functions for solving minimization
problems of the form.

    $argmin_x { d(x,w) + beta * g(x) }$,
where d is a metric, g is a "simple" function, and beta is a parameter controlling the trade-off between d and g.

TODO:
    - Add proximal operator for L2-squared penalty so that we can support this using R-FISTA.
    - Clean-up group-by-feature to be less ugly.
    - Should we retain the Orthant and GroupL1Orthant operators? They don't have a use at the moment.
"""

from typing import Optional
import math

import lab


class ProximalOperator:

    """Base class for proximal operators."""

    def __call__(self, w: lab.Tensor, beta: Optional[float] = None) -> lab.Tensor:
        """Evaluate the proximal_operator.

        :param w: parameters to which apply the operator will be applied.
        :param beta: the coefficient in the proximal operator. This is usually a step-size.
        :returns: prox(w)
        """

        raise NotImplementedError("A proximal operator must implement '__call__'!")


class Identity(ProximalOperator):
    """The proximal-operator for the zero-function.

    This proximal-operator always returns the input point.
    """

    def __call__(self, w: lab.Tensor, beta: Optional[float] = None) -> lab.Tensor:
        """Evaluate the identity operator.

        Args:
            w: the point at which the evaluate the operator.
            beta: the step-size for the proximal step. NOT USED.

        Returns:
            w, the original input point.
        """

        return w


class Regularizer(ProximalOperator):

    """Base class for proximal operators based on regularizers.

    Attributes:
        lam: the regularization strength. This must be non-negative.
    """

    def __init__(self, lam: float):
        """Initialize the proximal operator.

        Args:
            lam: a non-negative parameter controlling the regularization strength.
        """
        self.lam = lam


class L2(Regularizer):
    """The proximal operator for the squared l2-norm.

    The proximal operator returns the unique solution to the following optimization problem:

    .. math:: \\min_x \\{\\|x - w\\|_2^2 + \\frac{\\beta * \\lambda}{2} \\|x\\|_2^2\\}.

    The solution is the shrinkage operator :math:`x^* = (1 + \\beta)^{-1} w`.
    """

    def __call__(self, w: lab.Tensor, beta: float) -> lab.Tensor:
        """Evaluate the proximal operator given a point and a step-size.

        Args:
            w: the point at which the evaluate the operator.
            beta: the step-size for the proximal step.

        Returns:
            prox(w), the result of the proximal operator.
        """

        return w / (1 + beta * self.lam)


class L1(Regularizer):

    """The proximal operator for the l1-norm.

    The l1 proximal operator is sometimes known as the soft-thresholding operator and
    is the unique solution to the following problem:

    .. math:: \\min_x \\{\\|x - w\\|_2^2 + \\beta * \\lambda \\|x\\|_1\\}.
    """

    def __call__(self, w: lab.Tensor, beta: float) -> lab.Tensor:
        """Evaluate the proximal operator given a point and a step-size.

        Args:
            w: the point at which the evaluate the operator.
            beta: the step-size for the proximal step.

        Returns:
            prox(w), the result of the proximal operator.
        """

        return lab.sign(w) * lab.smax(lab.abs(w) - beta * self.lam, 0.0)


class GroupL1(Regularizer):

    """The proximal operator for the group-l1 regularizer.

    Given group indices :math:`\\calI`, the group-l1 regularizer is the sum of
    l2-norms of the groups, :math:`r(w) = \\sum_{i \\in \\calI} \\|w_i\\|_2`.
    The proximal operator is thus the unique solution to the following problem,

    .. math:: \\min_x \\{\\|x - w\\|_2^2 + \\beta * \\lambda \\sum_{i=1 \\in \\calI} \\|x_i\\|_2\\}.

    Groups are either defined to be the last axis of the point w.
    """

    def __call__(self, w: lab.Tensor, beta: float) -> lab.Tensor:
        """Evaluate the proximal operator given a point and a step-size.

        Args:
            w: the point at which the evaluate the operator.
            beta: the step-size for the proximal step.

        Returns:
            prox(w), the result of the proximal operator.
        """
        # compute the squared norms of each group.
        norms = lab.sqrt(lab.sum(w**2, axis=-1, keepdims=True))

        w_plus = lab.multiply(
            lab.safe_divide(w, norms), lab.smax(norms - self.lam * beta, 0)
        )

        return w_plus


class FeatureGroupL1(Regularizer):

    """The proximal operator for the feature-wise group-l1 regularizer.

    The feature-wise group-l1 regularizer is the sum of l2-norms of the
    weights for each feature,

    ..math:: r(w) = \\sum_{i \\in \\calI} \\|w_i\\|_2.

    The proximal operator is the unique solution to the following problem,

    .. math:: \\min_x \\{\\|x - w\\|_2^2 + \\beta * \\lambda
        \\sum_{j=1}^d \\|x_{:,j}\\|_2\\},

    where :math:`x_{:, j}` is the vector of all weights corresponding to
    feature :math:`j`.
    """

    def __call__(self, w: lab.Tensor, beta: float) -> lab.Tensor:
        """Evaluate the proximal operator given a point and a step-size.

        Args:
            w: the point at which the evaluate the operator.
            beta: the step-size for the proximal step.

        Returns:
            prox(w), the result of the proximal operator.
        """
        # compute the squared norms of each feature group.
        scaling = math.sqrt(w.shape[-1])

        norms = lab.sqrt(
            lab.sum(
                w**2,
                axis=tuple(range(len(w.shape) - 1)),
                keepdims=True,
            )
        )

        w_plus = lab.multiply(
            lab.safe_divide(w, norms),
            lab.smax(norms - scaling * self.lam * beta, 0),
        )

        return w_plus


# ==============
# REFACTOR BELOW
# ==============


class Orthant(ProximalOperator):

    """The projection operator for the orthant constrain, A_i x >= 0, where A_i
    is a diagonal matrix with (A_i)_jk in {-1, 1}."""

    def __init__(self, A: lab.Tensor):
        """
        A: a matrix of sign patterns defining orthants on which to project.
            The diagonal A_i is stored as the i'th column of A.
        """
        self.A = A
        if len(A.shape) == 3:
            self.sum_string = "ikj,imjk->imjk"
        else:
            self.sum_string = "kj,imjk->imjk"

    def __call__(self, w: lab.Tensor, beta: Optional[float] = None) -> lab.Tensor:
        """
        w: parameters to which the projection will be applied.
            This should be a (k x c x p x d) array, where each element of axis -1 corresponds to one column of A.
        beta: NOT USED. The coefficient in the proximal operator. This is usually a step-size.
        :returns: updated parameters.
        """

        return lab.where(
            lab.einsum(self.sum_string, self.A, w) >= 0, w, lab.zeros_like(w)
        )


class GroupL1Orthant(Regularizer):

    """Mixed group L1 penalty with orthant constraint."""

    def __init__(self, d: int, lam: float, A: lab.Tensor):
        """
        d: the dimensionality of each group for the regularizer.
        lam: the strength of the group-L1 regularizer.
        A: a matrix of sign patterns defining orthants on which to project.
            The diagonal A_i is stored as the i'th column of A.
        """
        self.d = d
        self.A = A
        self.lam = lam

        self.group_prox = GroupL1(lam)
        self.orthant_proj = Orthant(A)

    def __call__(self, w: lab.Tensor, beta: float) -> lab.Tensor:
        """
        w: parameters to which the projection will be applied.
            This should be a (k x c x p x d + n) array.
            The first 'd' entries in element of axis -1 correspond to the model weights (with group L1 regularizer)
            and the remaining 'n' entries correspond to the slack variables (with orthant constraint).
        beta: The coefficient in the proximal operator. This is usually a step-size.
        :returns: updated parameters.
        """
        model_weights, slacks = w[:, :, :, : self.d], w[:, :, :, self.d :]
        w_plus, s_plus = self.group_prox(model_weights, beta), self.orthant_proj(
            slacks, beta
        )

        return lab.concatenate([w_plus, s_plus], axis=-1)



"""Implementation of logistic regression for binary classification."""
from typing import Optional

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer
import scnn.private.loss_functions as loss_fns


class LogisticRegression(Model):

    """Logistic regression with for binary classification."""

    def __init__(self, d: int, regularizer: Optional[Regularizer] = None):
        """
        :param d: the dimensionality of the dataset (ie.number of features).
        :param regularizer: (optional) a penalty function controlling the flexibility of the model.
        """
        super().__init__(regularizer)

        self.weights = lab.zeros(d)

    def _forward(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        **kwargs,
    ) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the forward pass.
        """

        return loss_fns.logistic_fn(X @ w)

    def _logits(self, X: lab.Tensor, w: lab.Tensor) -> lab.Tensor:
        return X @ w

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute logistic objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the objective.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: objective L(f, (X,y)).
        """
        return loss_fns.logistic_loss(self._logits(X, w), y) / self._scaling(y, scaling)

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the logistic objective with respect to the
        model parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the gradient.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        logits = self._logits(X, w)
        return (
            -X.T
            @ lab.multiply(y, loss_fns.logistic_fn(-y * logits))
            / self._scaling(y, scaling)
        )



"""Group L1 Regularizer."""
from typing import Optional
import math

import lab

from scnn.private.models.regularizers.regularizer import Regularizer


class FeatureGroupL1Regularizer(Regularizer):

    """The group-sparsity inducing Group L1-regularizer over features.

    The group-L1 regularizer, sometimes called the L1-L2 regularizer,
    has the mathematical form,

        ::math.. R(w) = \\lambda \\sum_{i \\in \\mathcal{I} ||w_i||_2,

    where :math:`\\mathcal{I}` is collection of disjoint index sets specifying
    the groups for the regularizer. This class expects the second-to-last axis
    of the inputs/weights :math:`w` to be the feature-wise axis defining the
    different groups.

    This regularizer induces group-sparsity, meaning entire features will be
    dropped out of the model when :math:`\\lambda` is sufficiently large.

    Attributes:
        lam: the regularization strength.

    """

    def __init__(self, lam: float):
        """
        Args:
            lam: a tuning parameter controlling the regularization strength.
        """

        self.lam = lam

    def penalty(
        self,
        w: lab.Tensor,
        **kwargs,
    ) -> float:
        """Compute the penalty associated with the regularizer.

        Args:
            w: parameter at which to compute the penalty.

        Returns:
            The value of the penalty at w.
        """
        scaling = math.sqrt(w.shape[-1])

        feature_norms = (
            scaling
            * self.lam
            * lab.sqrt(
                lab.sum(
                    w ** 2,
                    axis=tuple(range(len(w.shape) - 1)),
                )
            )
        )

        return lab.sum(feature_norms)

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the minimum-norm subgradient (aka, the pseudo-gradient).

        :param w: parameter at which to compute the penalty gradient.
        :param base_grad: the gradient of the un-regularized objective. This is required
            to compute the minimum-norm subgradient.
        :returns: minimum-norm subgradient.
        """
        # requires base_grad to compute minimum-norm subgradient
        assert base_grad is not None
        scaling = math.sqrt(w.shape[-1])

        dims_to_sum = tuple(range(len(w.shape) - 1))
        weight_norms = lab.sqrt(
            lab.sum(
                w ** 2,
                axis=dims_to_sum,
                keepdims=True,
            )
        )
        grad_norms = lab.sqrt(
            lab.sum(
                base_grad ** 2,
                axis=dims_to_sum,
                keepdims=True,
            )
        )

        non_smooth_term = (
            base_grad
            * lab.smin(scaling * self.lam / grad_norms, 1)
            * (weight_norms == 0)
        )
        smooth_term = scaling * self.lam * lab.safe_divide(w, weight_norms)

        # match input shape
        subgrad = smooth_term - non_smooth_term

        return subgrad



"""Generate activation patterns for convex reformulations of neural networks.

Overview:
    This module provides functions for generating gate vectors and activation
    patterns for ReLU or threshold activation functions. These patterns are
    used when forming (subsampled) convex reformulations of neural networks
    with those activations. Gate vectors are also required for Gated ReLU
    networks. 

    An activation pattern is a vector of the form,

    .. math:: d_i = 1(X w \\geq 0),

    where :math:`1(z \\geq 0)` is an element-wise indicator function whose
    i'th element is one when :math:`z_i \\geq 0` and zero-otherwise and
    :math:`w \\in \\mathbb{R}^d` is a "gate vector".
    Forming the convex reformulation of a neural network with ReLU activations
    requires enumerating the activation patterns a single ReLU or threshold
    neuron can take on,

    .. math:: \\mathcal{D} = \\left\\{  d = 1(X w \\geq 0) : w \\in
        \\mathbb{R}^d \\right\\}.

    In practice, :math`\\mathcal{D}` can approximated sampling vectors
    :math:`w \\sim P` according to some distribution :math:`P` and then 
    computing the corresponding pattern :math:`d_i`.

TODO:
    - Where to put ugly conv mask code?
"""

from typing import Tuple, List, Literal, Optional
import math
from itertools import combinations

import numpy as np
from numpy.random import Generator


GateType = Literal[
    "dense",
    "feature_sparse",
    "convolutional",
]


def sample_gate_vectors(
    seed: int,
    d: int,
    n_samples: int,
    gate_type: GateType = "dense",
    order: Optional[int] = None,
) -> np.ndarray:
    """Generate gate vectors by random sampling.

    Args:
        seed: the random seed to use when generating the gates.
        d: the dimensionality of the gate vectors.
        n_samples: the number of samples to use.
        gate_type: the type of gates to sample. Must be one of:

            - `'dense'`: sample dense gate vectors.

            - `'feature_sparse'`: sample gate vectors which are sparse in
                specific features.

        order: the maximum order of feature sparsity to consider.
            Only used for `gate_type='feature_sparse'`. See
            :func:`sample_sparse_gates` for more details.

    Notes:
        It is possible to obtain more than `n_samples` gate vectors when
            `gate_type='feature_sparse'`
        if `len(sparsity_indices)` does not divide evenly into n_samples. In
            this case, we use the ceiling function to avoid sampling zero
            gates for some sparsity patterns.
    """

    rng = np.random.default_rng(seed)

    if gate_type == "dense":
        G = sample_dense_gates(rng, d, n_samples)
    elif gate_type == "feature_sparse":
        if order is None:
            raise ValueError(
                "`Order` must be specified when sampling feature-sparse gate."
            )
        sparsity_indices = generate_index_lists(d, order)
        G = sample_sparse_gates(rng, d, n_samples, sparsity_indices)
    elif gate_type == "convolutional":
        G = sample_convolutional_gates(rng, d, n_samples)
    else:
        raise ValueError(f"Gate type {gate_type} not recognized.")

    return G


def sample_dense_gates(
    rng: np.random.Generator,
    d: int,
    n_samples: int,
) -> np.ndarray:
    """Generate dense gate vectors by random sampling.

    Args:
        rng: a NumPy random number generator.
        d: the dimensionality of the gate vectors.
        n_samples: the number of samples to use.

    Returns:
        G -  a :math:`d \\times \\text{n_samples}` matrix of gate vectors.
    """
    G = rng.standard_normal((d, n_samples))

    return G


def sample_sparse_gates(
    rng: np.random.Generator,
    d: int,
    n_samples: int,
    sparsity_indices: List[List[int]],
) -> np.ndarray:
    """Generate feature-sparse gate vectors by random sampling.

    Args:
        rng: a NumPy random number generator.
        d: the dimensionality of the gate vectors.
        n_samples: the number of samples to use.
        sparsity_indices: lists of indices (i.e. features) for which sparse
            gates should be generated. Each index list will get
            `n_samples / len(sparsity_indices)` gates which are sparse in
            every feature except the given indices.

    Notes:
        - It is possible to obtain more than `n_samples` gate vectors if
            `len(sparsity_indices)` does not divide evenly into n_samples.
            In this case, we use the ceiling to avoid sampling zero gates for
            some sparsity patterns.

    Returns:
        G -  a :math:`d \\times \\text{n_samples}` matrix of gate vectors.
    """
    samples_per_list = math.ceil(n_samples / len(sparsity_indices))
    gates = []

    for indices in sparsity_indices:
        # create mask
        mask = np.zeros((d, 1))
        mask[indices, :] = 1

        G = sample_dense_gates(rng, d, samples_per_list)

        # project gates onto subspace
        gates.append(np.multiply(G, mask))

    return np.concatenate(gates, axis=1)


def sample_convolutional_gates(
    rng: np.random.Generator,
    d: int,
    n_samples: int,
) -> np.ndarray:
    """Generate convolutional gate vectors by random sampling.

    Args:
        rng: a NumPy random number generator.
        d: the dimensionality of the gate vectors.
        n_samples: the number of samples to use.

    Returns:
        G -  a :math:`d \\times \\text{n_samples}` matrix of gate vectors.
    """
    G = sample_dense_gates(rng, d, n_samples)

    if d == 784:
        conv_masks = _generate_conv_masks(rng, n_samples, 28, 1)
    elif d == 3072:
        conv_masks = _generate_conv_masks(rng, n_samples, 32, 3)
    else:
        assert (
            False
        ), "Convolutional patterns only implemented for MNIST or CIFAR datasets"

    G = G * conv_masks
    return G


def compute_activation_patterns(
    X: np.ndarray,
    G: np.ndarray,
    filter_duplicates: bool = True,
    filter_zero: bool = True,
    bias: bool = False,
    active_proportion: Optional[float] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute activation patterns corresponding to a set of gate vectors.

    Args:
        X: an (n x d) data matrix, where the examples are stored as rows.
        G: an (d x m) matrix of "gate vectors" used to generate the activation
            patterns.
        filter_duplicates: whether or not to remove duplicate activation
            patterns and the corresponding
        filter_zero: whether or not to filter the zero activation pattern and
            corresponding gates. Defaults to `True`.
        bias: whether or not a bias should be included in the gate vectors.
        active_proportion: force each gate to be active for
            `active_proportion`*n of the training examples using a bias
            term. This feature is only supported when `bias == True`.

    Returns:
        - `D`, an (n x p) matrix of (possibly unique) activation patterns where
            :math:`p \\leq m`

        - `G`, a (d x b) matrix of gate vectors generating `D`.
    """
    n, d = X.shape

    # need to extend the gates with zeros.
    if bias and G.shape[0] + 1 == X.shape[1]:
        G = np.concatenate([G, np.zeros((1, G.shape[1]))], axis=0)

    XG = np.matmul(X, G)

    if active_proportion is not None:
        # Gates must be augmented with a row of zeros to be valid.
        assert np.all(G[-1] == 0)
        # X must be augmented with a column of ones to be valid.
        assert np.all(X[:, -1] == X[0, -1])

        # set bias terms in G
        quantiles = np.quantile(XG, q=1 - active_proportion, axis=0, keepdims=True)
        XG = XG - quantiles
        G = G.copy()
        G[-1] = -np.ravel(quantiles)

    XG = np.maximum(XG, 0)
    XG[XG > 0] = 1

    if filter_duplicates:
        D, indices = np.unique(XG, axis=1, return_index=True)
        G = G[:, indices]

    # filter out the zero column.
    if filter_zero:
        non_zero_cols = np.logical_not(np.all(D == np.zeros((n, 1)), axis=0))
        D = D[:, non_zero_cols]
        G = G[:, non_zero_cols]

    return D, G


def generate_index_lists(
    d: int,
    order: int,
) -> List[List[int]]:
    """Generate lists of all groups of indices of order up to and including
    `order`.

    Args:
        d: the dimensionality or maximum index.
        order: the order to which the lists should be generated.
            For example, `order=2` will generate all singleton lists and
            all lists of pairs of indices.

    Returns:
        List of list of indices.
    """
    assert order > 0

    index_lists: List[List[int]] = []
    all_indices = list(range(d))

    for i in range(1, order + 1):
        d_choose_i = [list(comb) for comb in combinations(all_indices, i)]
        index_lists = index_lists + d_choose_i

    return index_lists


def _generate_conv_masks(
    rng: Generator,
    num_samples: int,
    image_size: int = 32,
    channels: int = 3,
    kernel_size: int = 3,
):
    upper_left_coords = rng.integers(
        low=0, high=image_size - kernel_size - 1, size=(num_samples, 2)
    )
    upper_left_indices = image_size * upper_left_coords[:, 0] + upper_left_coords[:, 1]
    upper_rows = [
        np.arange(upper_left_indices[i], upper_left_indices[i] + kernel_size)
        for i in range(num_samples)
    ]
    first_patch = [
        np.concatenate(
            [
                np.arange(
                    upper_rows[i][j],
                    upper_rows[i][j] + kernel_size * image_size,
                    image_size,
                )
                for j in range(kernel_size)
            ]
        )
        for i in range(num_samples)
    ]
    all_patches = [
        np.concatenate(
            [
                np.arange(
                    first_patch[i][j],
                    first_patch[i][j] + channels * image_size**2,
                    image_size**2,
                )
                for j in range(kernel_size**2)
            ]
        ).tolist()
        for i in range(num_samples)
    ]
    mask = np.zeros((num_samples, channels * image_size**2))

    def _transpose(x):
        dims = np.arange(len(x.shape))
        dims[[0, 1]] = [1, 0]

        return np.transpose(x, dims)

    mask[np.arange(num_samples), _transpose(all_patches)] = 1.0

    return mask.t()



"""Orthant constraint."""
from typing import Optional

import lab

from scnn.private.models.regularizers.regularizer import Regularizer

from scnn.private.prox import ProximalOperator


class Constraint(Regularizer):

    """Representation of a regularizer based on convex constraints of the form,
    f(W) geq 0, where f is a convex function and W are the model weights. In
    essence, this class is a wrapper for computation of the gradient mapping,

        g(xk) = xk - proj(xk - grad(xk)),
    which can be used as a stopping criterion.
    """

    lam = 0.0
    projection_op: ProximalOperator

    def __init__(self):
        raise NotImplementedError(
            "A constraint-based regularizer must implement its own constructor."
        )

    def _gradient_mapping(
        self,
        w: lab.Tensor,
        grad: lab.Tensor,
        step_size: float = 1.0,
    ) -> lab.Tensor:
        w_plus = w - step_size * grad
        return (w - self.projection_op(w_plus)) / step_size

    def penalty(self, w: lab.Tensor, **kwargs) -> float:
        """Compute the penalty associated with the regularizer.

        :param w: parameter at which to compute the penalty.
        :returns: penalty value
        """
        return 0.0

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        step_size: float = 1.0,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the objective with respect to the model
        parameters.

        :param w: parameter at which to compute the penalty gradient.
        :param base_grad: gradient from the base model.
        :param step_size: step-size to use when computing the gradient mapping.
        :returns: the gradient
        """
        assert base_grad is not None
        return self._gradient_mapping(w, base_grad, step_size) - base_grad



"""Block-diagonal matrix."""
from typing import Callable, Optional, Tuple, List

import numpy as np

from scipy.sparse.linalg import LinearOperator  # type: ignore

import lab


class MatVecOperator(LinearOperator):
    """Implementation of a generalized linear operator for matrix-vector
    products.

    Matrix-matrix products are implemented by naive iterative over columns.
    Note: use 'dot' to support composition of MatVecOperators.
    """

    def __init__(
        self,
        shape: Tuple[int, ...],
        forward: Callable[[lab.Tensor], lab.Tensor],
        transpose: Optional[Callable[[lab.Tensor], lab.Tensor]] = None,
    ):
        """
        :param shape: the shape of the linear operator.
        :param forward: the function which computes matrix-vector products for the matrix.
        :param transpose: the function which computes transposed matrix products.
        """
        self.forward = forward
        self.transpose = transpose
        self.shape = shape
        try:
            self.dtype = np.dtype(lab.get_dtype())
        except:
            self.dtype = lab.get_dtype()

    def matvec(self, v: lab.Tensor) -> lab.Tensor:
        return self.forward(v)

    def _matvec(self, v: lab.Tensor):
        return self.forward(v)

    def rmatvec(self, v: lab.Tensor) -> lab.Tensor:
        if self.transpose is None:
            raise ValueError(
                "MatVec instance cannot compute transpose products because it was not provided with a transpose operator."
            )

        return self.transpose(v)

    def matmat(self, V: lab.Tensor) -> lab.Tensor:
        return lab.stack([self.matvec(col) for col in V.T], axis=1)

    def _rmatvec(self, v: lab.Tensor):
        if self.transpose is None:
            raise ValueError(
                "MatVec instance cannot compute transpose products because it was not provided with a transpose operator."
            )

        return self.transpose(v)

    def rmatmat(self, V: lab.Tensor) -> lab.Tensor:
        if self.transpose is None:
            raise ValueError(
                "MatVec instance cannot compute transpose products because it was not provided with a transpose operator."
            )

        return lab.stack([self.rmatvec(col) for col in V.T], axis=1)



"""Convex formulations of neural networks."""

from typing import Optional, List, Tuple, Dict
from math import ceil

from scipy.sparse.linalg import LinearOperator  # type: ignore

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer
from . import operators
from scnn.private.utils import MatVecOperator
from scnn.private.loss_functions import squared_error, relu

# two-layer MLPs with ReLU activations.


class ConvexMLP(Model):
    """Convex formulation of a two-layer neural network (multi-layer
    perceptron) with ReLU activations."""

    def __init__(
        self,
        d: int,
        D: lab.Tensor,
        U: lab.Tensor,
        kernel: str = operators.EINSUM,
        regularizer: Optional[Regularizer] = None,
        c: int = 1,
    ) -> None:
        """
        :param d: the dimensionality of the dataset (ie. number of features).
        :param D: array of possible sign patterns.
        :param U: array of hyperplanes creating the sign patterns.
        :param kernel: the kernel to drive the matrix-vector operations.
        """
        super().__init__(regularizer)

        self.d = d
        self.p = D.shape[1]  # each column is a unique sign pattern
        self.c = c
        self.weights = lab.zeros(
            (c, self.p, self.d)
        )  # one linear model per sign pattern

        self.D = D
        self.U = U
        self.kernel = kernel

        (
            self._data_mvp,
            self._data_t_mvp,
            self._gradient,
            self._hessian_mvp,
            self._bd_hessian_mvp,
        ) = operators.get_kernel(kernel)

        (
            self._data_builder,
            self._hessian_builder,
            self._bd_hessian_builder,
        ) = operators.get_matrix_builders(kernel)

        self._train = True

    def _weights(self, w: Optional[lab.Tensor]) -> lab.Tensor:
        return self.weights if w is None else w

    def get_reduced_weights(self) -> lab.Tensor:
        return self.weights

    def _signs(self, X: lab.Tensor, D: Optional[lab.Tensor] = None):
        local_D = self.D
        if D is not None:
            return D
        elif not self._train:
            local_D = relu(X @ self.U)
            local_D[local_D > 0] = 1

        return local_D

    def _forward(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the forward pass.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :returns: predictions for X.
        """
        return self._data_mvp(w, X, self._signs(X, D))

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute the l2 objective with respect to the model weights.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the objective.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the objective
        """

        return squared_error(self._forward(X, w, D), y) / (
            2 * self._scaling(y, scaling)
        )

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        flatten: bool = False,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the l2 objective with respect to the model
        weights. As in 'self.__call__' above, we could optimize this by
        implementing it in a faster low-level language.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the gradient.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :param flatten: whether or not to flatten the blocks of the gradient into a single vector.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        grad = self._gradient(w, X, y, self._signs(X, D)) / self._scaling(
            y, scaling
        )

        if flatten:
            grad = grad.reshape(-1)

        return grad

    def data_operator(
        self,
        X: lab.Tensor,
        D: Optional[lab.Tensor] = None,
    ) -> LinearOperator:
        """Construct a matrix operator to evaluate the matrix-vector equivalent
        to the sum,

        .. math::

            \\sum_i D_i X v_i

        where v_i is the i'th block of the input vector 'v'. This is equivalent to
        constructing the expanded matrix :math:`A = [D_1 X, D_2 X, ..., D_P X]` and then evaluating :math:`Av`.
        Use 'data_matrix' to directly compute $A$.
        :param X: (n,d) array containing the data examples.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :returns: LinearOperator which computes the desired product.
        """
        n, _ = X.shape
        pd = self.d * self.p
        local_D = self._signs(X, D)

        # pre-compute extended matrix
        if self.kernel == operators.DIRECT:
            expanded_X = self._data_builder(X, local_D)

            def forward(v):
                return lab.squeeze(
                    self._data_mvp(v, X=X, D=local_D, expanded_X=expanded_X)
                )

            def transpose(v):
                return lab.squeeze(
                    self._data_t_mvp(v, X=X, D=local_D, expanded_X=expanded_X)
                )

        else:

            def forward(v):
                return lab.squeeze(self._data_mvp(v, X=X, D=local_D))

            def transpose(v):
                return lab.squeeze(self._data_t_mvp(v, X=X, D=local_D))

        op = MatVecOperator(
            shape=(n, pd), forward=forward, transpose=transpose
        )

        return op

    def add_new_patterns(
        self, patterns: lab.Tensor, weights: lab.Tensor, remove_zero=False
    ) -> lab.Tensor:
        """Attempt to augment the current model with additional sign patterns.

        :param patterns: the tensor of sign patterns to add to the current set.
        :param weights: the tensor of weights which induced the new patterns.
        :param remove_zero: whether or not to remove the zero vector.
        :returns: None
        """

        if lab.size(patterns) > 0:
            # update neuron count.
            self.D = lab.concatenate([self.D, patterns], axis=1)
            self.U = lab.concatenate([self.U, weights.T], axis=1)

            # initialize new model components at 0.
            added_weights = lab.zeros((self.c, weights.shape[0], self.d))
            self.weights = lab.concatenate(
                [self.weights, added_weights], axis=1
            )

            # filter out the zero column.
            if remove_zero:
                non_zero_cols = lab.logical_not(
                    lab.all(self.D == lab.zeros((self.D.shape[0], 1)), axis=0)
                )
                self.D = self.D[:, non_zero_cols]
                self.U = self.U[:, non_zero_cols]
                self.weights = self.weights[:, non_zero_cols]

            self.p = self.D.shape[1]

        return added_weights

    def batch_X(
        self, batch_size: Optional[int], X: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:
        D = self._signs(X)

        if batch_size is None:
            return [{"X": X, "D": D}]

        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {
                "X": X[i * batch_size : (i + 1) * batch_size],
                "D": D[i * batch_size : (i + 1) * batch_size],
            }
            for i in range(n_batches)
        ]

    def batch_Xy(
        self, batch_size: Optional[int], X: lab.Tensor, y: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:
        D = self._signs(X)

        if batch_size is None:
            return [{"X": X, "y": y, "D": D}]

        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {
                "X": X[i * batch_size : (i + 1) * batch_size],
                "y": y[i * batch_size : (i + 1) * batch_size],
                "D": D[i * batch_size : (i + 1) * batch_size],
            }
            for i in range(n_batches)
        ]

    def sign_patterns(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> Tuple[lab.Tensor, lab.Tensor]:
        """Compute the gradient of the l2 objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: (optional) specific parameter at which to compute the sign patterns.
        :returns: the set of sign patterns active at w or the current models parameters if w is None.
        """

        return lab.sign(relu(X @ self.U)), self.U.T



"""Implementation of one-layer ReLU MLP with squared-error objective."""
from typing import Optional, Tuple

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer
from scnn.private.loss_functions import squared_error, relu


class ReLUMLP(Model):

    """One-layer ReLU MLP with squared-error objective."""

    activation_history: Optional[lab.Tensor] = None
    weight_history: Optional[lab.Tensor] = None

    def __init__(
        self,
        d: int,
        p: int,
        regularizer: Optional[Regularizer] = None,
        c: int = 1,
    ):
        """
        :param d: the dimensionality of the dataset (ie.number of features).
        :param p: the number of hidden units.
        :param regularizer: (optional) a penalty function controlling the flexibility of the model.
        """
        super().__init__(regularizer)
        self.d = d
        self.p = p
        self.c = c

        # use random initialization by default.
        self.weights = lab.tensor(
            lab.np_rng.standard_normal((d * p + p * c)), dtype=lab.get_dtype()
        )

    def set_weights(self, weights: lab.Tensor):
        # weights include second layer
        if weights.shape == (self.d * self.p + self.p * self.c):
            self.weights = weights
        # weights consist only of first-layer weights.
        elif weights.shape == (self.d * self.p,):
            self.weights[: self.p * self.d] = weights
        # weights consist only of second-layer weights.
        elif weights.shape == (self.p * self.c,):
            self.weights[self.p * self.d :] = weights
        # weights need to be flattened.
        elif lab.size(weights) == (self.d * self.p + self.p * self.c):
            self.weights = lab.ravel(weights)
        else:
            raise ValueError(
                f"Weights with shape {weights.shape} cannot be set to ReluMLP with weight shape {self.d * self.p + p}."
            )

    def _split_weights(self, w: lab.Tensor) -> Tuple[lab.Tensor, lab.Tensor]:

        return (
            w[: self.d * self.p].reshape(self.p, self.d),
            w[self.d * self.p :].reshape(self.c, self.p),
        )

    def _join_weights(self, w1: lab.Tensor, w2: lab.Tensor) -> lab.Tensor:

        return lab.concatenate([lab.ravel(w1), lab.ravel(w2)])

    def _forward(self, X: lab.Tensor, w: lab.Tensor, **kwargs) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the forward pass.
        """
        w1, w2 = self._split_weights(w)

        return relu(X @ w1.T) @ w2.T

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: specific parameter at which to compute the forward pass.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: objective L(f, (X, y)).
        """
        return squared_error(self._forward(X, w), y) / (2 * self._scaling(y, scaling))

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the l2 objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: parameter at which to compute the gradient pass.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        w1, w2 = self._split_weights(w)
        Z = relu(X @ w1.T)
        D = lab.sign(Z)

        residuals = Z @ w2.T - y

        return self._grad_helper(X, residuals, w1, w2, D, Z) / self._scaling(y, scaling)

    def _grad_helper(
        self,
        X: lab.Tensor,
        residuals: lab.Tensor,
        w1: lab.Tensor,
        w2: lab.Tensor,
        D: lab.Tensor,
        Z: lab.Tensor,
    ) -> lab.Tensor:

        # always uses the 0 subgradient of max(x, 0) when x = 0.
        g1 = lab.einsum("ij, il, ik, lj->jk", D, residuals, X, w2)
        g2 = residuals.T @ Z

        return self._join_weights(g1, g2)

    def sign_patterns(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> Tuple[lab.Tensor, lab.Tensor]:
        """Compute the gradient of the l2 objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: (optional) specific parameter at which to compute the sign patterns.
        :returns: the set of sign patterns active at w or the current models parameters if w is None.
        """

        w1, w2 = self._split_weights(self._weights(w))
        return lab.sign(lab.smax(lab.matmul(X, w1.T), 0)), w1



"""Optimization methods for training neural networks by convex reformulation.

Notes:
    We only support the squared loss at the moment.

Todo:
    - Implement the cone decomposition optimizer for training ReLU model by
        (1) training a Gated ReLU model and (2) decomposing that model onto
        the conic difference.
"""
from typing import List

from .models import (
    Model,
    LinearModel,
    ConvexReLU,
    NonConvexReLU,
    ConvexGatedReLU,
    NonConvexGatedReLU,
)


class Optimizer:
    """Base class for optimizers.

    Attributes:
        model: the model that should be optimized.
    """

    def __init__(self, model: Model):
        """Initialize the optimizer.

        Args:
            model: the model to be optimized. It will be checked for
                compatibility with the optimizer.
        """
        self.model = model

    def cpu_only(self) -> bool:
        return False


class RFISTA(Optimizer):
    """Fast proximal-gradient solver for Gated ReLU models.

    This optimizer solves the convex Gated ReLU training problem by directly
    minimizing the convex reformulation,

    .. math:: F(u) = L\\left(\\sum_{D_i \\in \\mathcal{D}} D_i X u_{i}),
        y\\right) + \\lambda R(u),

    where :math:`L` is a convex loss function, :math:`R` is a regularizer,
    and :math:`\\lambda` is the regularization strength.

    Attributes:
        model: the model to be optimized.
        max_iters: the maximum number of iterations to run the solver.
        tol: the tolerance for terminating the optimization procedure early;
            `tol` will be checked against the squared norm of the minimum-norm
            subgradient.
    """

    def __init__(
        self, model: Model, max_iters: int = 10000, tol: float = 1e-6
    ):
        """
        Args:
            model: the model to be optimized. It will be checked for
                compatibility with the optimizer.
            max_iters: the maximum number of iterations to run the solver.
            tol: the tolerance for terminating the optimization procedure.
        """

        if not isinstance(model, (ConvexGatedReLU, LinearModel)):
            raise ValueError(
                "R-FISTA can only be used to train Gated ReLU models and \
                        linear model."
            )

        super().__init__(model)
        self.max_iters = max_iters
        self.tol = tol


class AL(Optimizer):
    """Augmented Lagrangian (AL) method for ReLU model.

    This optimizer solves the convex ReLU training problem by forming the
    "augmented Lagrangian",

    .. math:: \\mathcal{L}(v, w, \\gamma, \\xi) = F(v,w) + \\delta \\sum_{D_i}
        \\left[\\|(\\frac{\\gamma_i}{\\delta} - (2D_i - I)X v_i)_+\\|_2^2 +
        \\|(\\frac{\\xi_i}{\\delta} - (2D_i - I)X v_i)_+\\|_2^2 \\right],

    where :math:`\\delta > 0` is the penalty strength, :math:`(\\gamma, \\xi)`
    are the dual parameters, and

    .. math:: F(v,w) = L\\left(\\sum_{D_i \\in \\mathcal{D}} D_i X (v_{i} -
        w_{i}), y\\right) + \\lambda R(v, w),

    is the regularized training loss.

    The AL method alternates between the "primal" problem of minimizing
    :math:`\\mathcal{L}(v, w, \\gamma, \\xi)` in terms of :math:`v, w` and
    the updates to the dual parameters, :math:`\\gamma, \\xi`.
    This procedure will return an primal-dual optimal pair
    :math:`(v,w), (\\gamma, \\xi)` in the (dual) iteration limit.

    The solver will terminate early if an approximate KKT point is computed.

    Attributes:
        model: the model to be optimized.
        max_primal_iters: the maximum number of iterations to run the primal
            optimization method before exiting.
        max_dual_iters: the maximum number of dual updates that will be
            performed before exiting.
        tol: the maximum squared l2-norm of the gradient of the Lagrangian
            function for the KKT conditions to be considered approximately
            satisfied.
        constraint_tol: the maximum violation of the constraints permitted
            for the KKT conditions to be considered approximately
            satisfied.
        delta: the initial penalty strength.
    """

    def __init__(
        self,
        model: Model,
        max_primal_iters: int = 10000,
        max_dual_iters: int = 10000,
        tol: float = 1e-6,
        constraint_tol: float = 1e-6,
        delta: float = 1000,
    ):
        """
        Args:
            model: the model to be optimized. It will be checked for
                compatibility with the optimizer.
            max_primal_iters: the maximum number of iterations to run the
                primal optimization method before exiting.
            max_dual_iters: the maximum number of dual updates that will be
                performed before exiting.
            tol: the maximum squared l2-norm of the gradient of the Lagrangian
                function for the KKT conditions to be considered approximately
                satisfied.
            constraint_tol: the maximum violation of the constraints permitted
                for the KKT conditions to be considered approximately
                satisfied.
            delta: the initial penalty strength.
        """
        if not isinstance(model, (ConvexReLU)):
            raise ValueError(
                "The AL optimizer can only be used to train ReLU models."
            )

        super().__init__(model)
        self.max_primal_iters = max_primal_iters
        self.max_dual_iters = max_dual_iters
        self.tol = tol
        self.constraint_tol = constraint_tol
        self.delta = delta


class LeastSquaresSolver(Optimizer):
    """Direct solver for the unregularized or :math:`\\ell_2`-regularized
    Gated ReLU training problem.

    The Gated ReLU problem with no regularization or :math:`\\ell_2`
    regularization is

    .. math:: F(u) = \\frac{1}{2}\\|\\sum_{D_i \\in \\mathcal{D}} D_i X u_{i}
        - y\\|_2^2 + \\lambda \sum_{D_i \\in \\mathcal{D}} \\|u_i\\|_2^2.

    This is a convex quadratic problem equivalent to ridge regression which
    this optimizer solves using conjugate-gradient (CG) type methods.
    Either LSMR [1] or LSQR [2] can be used; implementations are provided by
    `scipy.sparse.linalg
    <https://docs.scipy.org/doc/scipy/reference/sparse.linalg.html>`_.
    See `SOL <https://web.stanford.edu/group/SOL/download.html>`_ for details
    on these methods.

    Attributes:
        model: the model that should be optimized.
        solver: the underlying CG-type solver to use.
        max_iters: the maximum number of iterations to run the optimization
            method.
        tol: the tolerance for terminating the optimization procedure early.

    Notes:
        - This solver only supports computation on CPU. The user's choice of
            backend will be overridden if necessary.

        - This solver only supports :class:`L2 <scnn.regularizers.L2>`
            regularization or no regularization.

    References:
        [1] D. C.-L. Fong and M. A. Saunders, LSMR: An iterative algorithm
            for sparse least-squares problems, SIAM J. Sci. Comput. 33:5,
            2950-2971, published electronically Oct 27, 2011.

        [2] C. C. Paige and M. A. Saunders, LSQR: An algorithm for sparse
        linear equations and sparse least squares, TOMS 8(1), 43-71 (1982)
    """

    def __init__(
        self,
        model: Model,
        solver: str = "lsmr",
        max_iters: int = 1000,
        tol=1e-6,
    ):
        """
        Args:
            model: the model to be optimized. It will be checked for
                compatibility with the optimizer.
            solver: the underlying CG-type solver to use.
                Must be one of "LSMR" or "LSQR".
            max_iters: the maximum number of iterations to run the CG-method.
            tol: the tolerance for terminating the optimization procedure
                early.
        """
        if not isinstance(model, (ConvexGatedReLU)):
            raise ValueError(
                "The least-squares solver only supports Gated ReLU models."
            )

        super().__init__(model)

        self.solver = solver
        self.max_iters = max_iters
        self.tol = tol

    def cpu_only(self) -> bool:
        return True


class CVXPYSolver(Optimizer):
    """Solve convex reformulations using `CVXPY <https://www.cvxpy.org>`_ as a
    interface to different interior-point solvers.

    `CVXPY <https://www.cvxpy.org>`_ provides a framework for denoting and
    solving convex optimization problems. The framework is compatible with a
    variety of solvers (mainly interior point methods); see
    `choosing a solver
    <https://www.cvxpy.org/tutorial/advanced/index.html#choosing-a-solver>`_
    for details. Note that some solvers may require additional libraries
    and/or licences to be installed.


    Attributes:
        model: the model that should be optimized.
        solver: the underlying solver to use with CVXPY.
        solver_kwargs: a dictionary of keyword arguments to be passed directly
            to the underlying solver.
        clean_sol: whether or not to clean the solution using a
            proximal-gradient step.
     Notes:
         - This solver only supports computation on CPU. The user's choice of
            device will be overridden if necessary.
    """

    supported_solvers: List[str] = ["ecos", "cvxopt", "scs", "gurobi", "mosek"]

    def __init__(
        self, model: Model, solver, solver_kwargs={}, clean_sol=False
    ):
        """
        Args:
            model: the model to be optimized. It will be checked for
                compatibility with the optimizer.
            solver: a string identifying the solver to use with CVXPY. Only
                'ecos', 'cvxopt', 'scs', 'gurobi', and 'mosek' are supported.
                See `choosing a solver
                <https://www.cvxpy.org/tutorial/advanced/index.html#choosing
                -a-solver>`_ for details.
            solver_kwargs: keyword arguments that will be passed directly to
                the underlying solver. See `solver options
                <https://www.cvxpy.org/tutorial/advanced/index.html#setting
                -solver-options>`_ for details.
            clean_sol: whether or not to clean the solution using a
                proximal-gradient step. This is only supported for Gated ReLU
                problems.

        """
        super().__init__(model)

        if clean_sol and isinstance(model, ConvexReLU):
            raise ValueError(
                "Cleaning solutions using a proximal-gradient step is \
                only supported for unconstrained problems."
            )

        if solver not in self.supported_solvers:
            raise ValueError(
                f"CVXPYSolver does not support {solver}; it only supports \
                {self.supported_solvers} for now."
            )

        self.solver = solver
        self.solver_kwargs = solver_kwargs
        self.clean_sol = clean_sol

    def cpu_only(self) -> bool:
        return True


class ExactConeDecomposition(Optimizer):
    """Two-step method for approximately optimizing ReLU models.

    ConeDecomposition first solves the Gated ReLU problem,

    .. math:: \\min_{u} L\\left(\\sum_{D_i \\in \\mathcal{D}} D_i X u_{i}),
        y\\right) + \\lambda R(u),

    and then decomposes the solution :math:`u^*` onto the Minkowski differences
    :math:`K_i - K_i` to approximate the ReLU training problem. The cone
    decomposition is solved exactly using interior-point methods and
    CVXPY...
    """

    def __init__(self):
        raise NotImplementedError(
            "Exact cone decomposition is not \
                implemented yet."
        )


class ApproximateConeDecomposition(Optimizer):
    """Two-step method for approximately optimizing ReLU models.

    ConeDecomposition first solves the Gated ReLU problem,

    .. math:: \\min_{u} L\\left(\\sum_{D_i \\in \\mathcal{D}} D_i X u_{i}),
        y\\right) + \\lambda R(u),

    and then decomposes the solution :math:`u^*` onto the Minkowski differences
    :math:`K_i - K_i` to approximate the ReLU training problem. The cone
    decomposition is approximated by solving

    .. math:: \\min_{w} \\sum_{D_i \\in \\mathcal{D}}\\left[
        \\| \\left(\\tilde X_i w - \\min\\{\\tilde X_i u^*, 0\\}\\right)_+\\|_2^2
        + \\rho \\|w_i\\|_2^2 \\right],

    where :math:`\\tilde X_i = (2D_i - I) X`.
    The regularization :math:`\\rho` controls the quality of the approximation;
    taking :math:`\\rho \\rightarrow 0` will return a feasible solution.
    A feasible solution is guaranteed to preserve the value of the loss
    :math:`L`, but can substantially blow-up the model norm. As such, it is
    only an approximation to the ReLU training problem when
    :math:`\\lambda > 0`.

    Attributes:
        model: the model that should be optimized.
        max_iters: the maximum number of iterations to run the R-FISTA sovler.
        tol: the tolerance for terminating R-FISTA early.
        rho: the strength of the penalty term in the decomposition program.
        d_max_iters: the maximum number of iterations for the decomposition
            program.
        d_tol: the tolerance for terminating the decomposition program.
    """

    def __init__(
        self,
        model: Model,
        max_iters: int = 10000,
        tol: float = 1e-6,
        rho: float = 1e-10,
        d_max_iters: int = 10000,
        d_tol: float = 1e-10,
    ):
        """Initialize the optimizer."""
        if not isinstance(model, (ConvexGatedReLU)):
            raise ValueError(
                "Cone decomposition methods are only compatible with \
                Gated ReLU models."
            )

        super().__init__(model)
        self.rfista = RFISTA(model, max_iters, tol)
        self.rho = rho
        self.d_max_iters = d_max_iters
        self.d_tol = d_tol



"""Optimization procedures."""

# ===== module exports ===== #

from .optimization_procedure import OptimizationProcedure, ITER_LOG_FREQ, EPOCH_LOG_FREQ, METRIC_FREQ
from .iterative import IterativeOptimizationProcedure
from .torch_loop import TorchLoop
from .double_loop_procedure import DoubleLoopProcedure


__all__ = [
    "OptimizationProcedure",
    "IterativeOptimizationProcedure",
    "TorchLoop",
    "DoubleLoopProcedure",
    "METRIC_FREQ",
    "ITER_LOG_FREQ",
    "EPOCH_LOG_FREQ",
]



""""Direct" operators for convex neural networks based on manually constructing
the underlying matrices."""
from typing import Optional

import lab

# operators


def data_mvp(
    v: lab.Tensor,
    X: lab.Tensor,
    D: lab.Tensor,
    expanded_X: Optional[lab.Tensor] = None,
    optimize: Optional[bool] = None,
) -> lab.Tensor:
    if expanded_X is None:
        expanded_X = expanded_data_matrix(X, D)
    w = v.reshape(-1, X.shape[1]*D.shape[1])
    return expanded_X @ w.T


def data_t_mvp(
    v: lab.Tensor,
    X: lab.Tensor,
    D: lab.Tensor,
    expanded_X: Optional[lab.Tensor] = None,
    optimize: Optional[bool] = None,
) -> lab.Tensor:
    if expanded_X is None:
        expanded_X = expanded_data_matrix(X, D)
    w = v.reshape(-1)
    return expanded_X.T @ w


def gradient(
    v: lab.Tensor,
    X: lab.Tensor,
    y: lab.Tensor,
    D: lab.Tensor,
    expanded_X: Optional[lab.Tensor] = None,
    optimize: Optional[bool] = None,
) -> lab.Tensor:
    if expanded_X is None:
        expanded_X = expanded_data_matrix(X, D)

    return (expanded_X.T @ (data_mvp(v, X, D, expanded_X) - y)).T.reshape(*v.shape)


def hessian_mvp(
    v: lab.Tensor,
    X: lab.Tensor,
    D: lab.Tensor,
    hessian: Optional[lab.Tensor] = None,
    optimize: Optional[bool] = None,
) -> lab.Tensor:

    if hessian is None:
        hessian = expanded_hessian(X, D)
        # flatten the Hessian.

        hessian = [entry for entry in hessian]
        hessian = lab.concatenate(hessian, axis=1)
        hessian = [entry for entry in hessian]
        hessian = lab.concatenate(hessian, axis=1)

    w = v.reshape(-1, X.shape[1]*D.shape[1])
    return (hessian @ w.T).T.reshape(*v.shape)


def bd_hessian_mvp(
    v: lab.Tensor,
    X: lab.Tensor,
    D: lab.Tensor,
    bd_hessian: Optional[lab.Tensor] = None,
    optimize: Optional[bool] = None,
) -> lab.Tensor:
    if bd_hessian is None:
        bd_hessian = expanded_bd_hessian(X, D)

    w = v.reshape(-1, D.shape[1], X.shape[1])
    res = []

    for i, block in enumerate(bd_hessian):
        res.append((block @ w[:, i].T).T)

    return lab.transpose(lab.stack(res), 0, 1).reshape(*v.shape)


# builders


def data(X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return expanded_data_matrix(X, D)


def hessian(X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return expanded_hessian(X, D)


def bd_hessian(X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return expanded_bd_hessian(X, D)


# helpers for direct matrix operations.


def expanded_data_matrix(
    X: lab.Tensor,
    D: lab.Tensor,
) -> lab.Tensor:
    """Construct the expanded data matrix A = [D_1 X, D_2 X, ..., D_p X] associated
    with the linear model,
        $sum_i D_i X v_i$
    where v_i is the i'th block of the input vector 'v'. Note: direct computation of A
    is not efficient; whenever possible, use the einsum kernels to construct an operator
    which can evaluate $A v$ and $A.T w$.
    :param X: (n,d) array containing the data examples.
    :param D: array of possible sign patterns.
    :returns: expanded data matrix .
    """
    P = D.shape[1]
    blocks = []
    for i in range(P):
        res = lab.diag(D[:, i]) @ X
        blocks.append(res)

    return lab.concatenate(blocks, axis=1)


def expanded_hessian(X: lab.Tensor, D: lab.Tensor, flatten: bool = False) -> lab.Tensor:
    """Construct Hessian of the squared loss for the linear model.

        $sum_i D_i X v_i$
    where v_i is the i'th block of the input vector 'v'. Note: direct computation
    is not efficient; whenever possible, use the einsum kernels to construct an operator
    which can evaluate $H v$ instead of calling this function.
    :param X: (n,d) array containing the data examples.
    :param D: array of possible sign patterns.
    :param flatten: whether or not to flatten the output matrix.
    :returns: Hessian of the expanded linear model.
    """

    P = D.shape[1]

    # compute matrix blocks by hand
    rows = []
    for i in range(P):
        blocks = []
        for j in range(P):
            H_ij = X.T @ lab.diag(D[:, i]) @ lab.diag(D[:, j]) @ X

            blocks.append(H_ij)

        rows.append(lab.stack(blocks))

    return lab.stack(rows)


def expanded_bd_hessian(
    X: lab.Tensor,
    D: lab.Tensor,
) -> lab.Tensor:
    """Construct the diagonal blocks H_i of the Hessian of the squared loss for
    the linear model.

        $sum_i D_i X v_i$
    where v_i is the i'th block of the input vector 'v'. Note: direct computation
    is not efficient; whenever possible, use the einsum kernels to construct an operator
    which can evaluate $H v$ instead of calling this function.
    :param X: (n,d) array containing the data examples.
    :param D: array of possible sign patterns.
    :returns: array with shape (P, d, d), where the first axis indexes the blocks.
    """
    P = D.shape[1]

    # compute matrix blocks by hand
    diagonal = []
    for i in range(P):
        H_ii = X.T @ lab.diag(D[:, i]) @ lab.diag(D[:, i]) @ X

        diagonal.append(H_ii)

    return lab.stack(diagonal)



"""Basic line-search loop."""

from typing import Callable, Dict, Tuple, Any

import lab
from scnn.private.methods.line_search import Backtracker, LSCondition

# constants

MAX_ATTEMPTS: int = 50
FAILURE_STEP: float = 1e-8


# line-search


def ls(
    test_fn: Callable,
    w: lab.Tensor,
    f0: float,
    descent_dir: lab.Tensor,
    grad: lab.Tensor,
    obj_fn: Callable,
    grad_fn: Callable,
    init_step_size: float,
    ls_cond: LSCondition,
    backtrack: Backtracker,
) -> Tuple[lab.Tensor, float, float, Dict[str, Any]]:
    """Take one step of a iterative method using a line-search to pick the
    step-size.

    :param test_fn: function that returns the next test point given w, grad, and the next step_size.
    :param w: the parameters to be updated.
    :param f0: the objective function evaluated at w.
    :param descent_dir: the descent direction for the step.
    :param grad: the gradient of the smooth component of the loss function.
    :param obj_fn: function that returns the objective when called at w.
    :param grad_fn: function that returns the gradient of the smooth component of the loss function.
    :param init_step_size: the first step-size to use.
    :param ls_cond: the line-search condition to check.
    :param backtrack: a rule for calculating the next step-size to try.
    :returns: (w_next, f1, step_size, exit_state) -- the updated parameters, objective value, new step-size, and exit state.
    """
    # setup
    success = True
    step_size = init_step_size

    # run line-search
    w_plus = test_fn(w, descent_dir, step_size)
    f1 = obj_fn(w_plus)
    attempts = 1

    while not ls_cond(f0, f1, w_plus - w, grad, step_size):
        step_size = backtrack(step_size)
        w_plus = test_fn(w, descent_dir, step_size)
        f1 = obj_fn(w_plus)
        attempts += 1

        # exceeded max attempts
        if attempts > MAX_ATTEMPTS:
            step_size = FAILURE_STEP
            w_plus = test_fn(w, grad, step_size)
            success = False
            break

    exit_state = {"attempts": attempts, "success": success, "step_size": step_size}

    return w_plus, f1, step_size, exit_state



"""End-to-end tests for solvers.

TODO:
    - Test cone decomposition method once implemented.
"""

import unittest

import numpy as np

from scnn.private.utils.data.synthetic import gen_regression_data
from scnn.regularizers import NeuronGL1, FeatureGL1, L2
from scnn.solvers import RFISTA, AL, LeastSquaresSolver, CVXPYSolver
from scnn.metrics import Metrics
from scnn.activations import sample_gate_vectors
from scnn.models import ConvexGatedReLU, ConvexReLU
from scnn.optimize import optimize_model


class TestRegularizers(unittest.TestCase):
    """Test solving convex reformulations with different regularizers."""

    def setUp(self):
        # Generate realizable synthetic classification problem (ie. Figure 1)
        n_train = 100
        n_test = 100
        d = 10
        self.c = 2
        kappa = 1
        self.max_neurons = 100

        (
            (self.X_train, self.y_train),
            (self.X_test, self.y_test),
            _,
        ) = gen_regression_data(
            123, n_train, n_test, d, self.c, kappa=kappa
        )
        self.lam = 0.0001

        # Instantiate convex model and other options.
        self.G = sample_gate_vectors(np.random.default_rng(123), d, self.max_neurons)
        self.metrics = Metrics()
        self.regularizer = NeuronGL1(0.01)

    def test_RFISTA(self):
        """Test the R-FISTA solver on a gated ReLU problem."""
        model = ConvexGatedReLU(self.G, c=self.c)
        solver = RFISTA(model, tol=1e-6)
        model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

    def test_AL(self):
        """Test the augmented Lagrangian solver on a ReLU problem."""
        model = ConvexReLU(self.G, c=self.c)

        # terminate early.
        solver = AL(model, tol=1e-3, constraint_tol=1e-3)
        model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

    def test_least_squares(self):
        """Test the least-squares solver on an l2-regularized gated ReLU problem."""
        model = ConvexGatedReLU(self.G, c=1)
        solver = LeastSquaresSolver(model)
        model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train[:, 0],  # only supports scalar output for now.
            self.X_test,
            self.y_test,
            L2(0.01),
        )

    def test_CVXPY(self):
        """Test CVXPY interior-point solvers on Gated ReLU and ReLU problems."""
        model = ConvexGatedReLU(self.G, c=self.c)
        solver = CVXPYSolver(model, solver="ecos")
        model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )

        model = ConvexReLU(self.G, c=self.c)
        solver = CVXPYSolver(model, solver="ecos")
        model, _ = optimize_model(
            model,
            solver,
            self.metrics,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            self.regularizer,
        )


if __name__ == "__main__":
    unittest.main()



"""Einsum-based operators for convex neural networks."""

import numpy as np

import lab


def data_mvp(v: lab.Tensor, X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    w = v.reshape(-1, D.shape[1], X.shape[1])
    return lab.einsum("ij, lkj, ik->il", X, w, D)


def data_t_mvp(v: lab.Tensor, X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return lab.einsum("ij, i, il->jl", D, v, X).reshape(-1)


# TODO: accept residual as a parameter and remove data_mvp from this operation.
def gradient(
    v: lab.Tensor,
    X: lab.Tensor,
    y: lab.Tensor,
    D: lab.Tensor,
) -> lab.Tensor:
    w = v.reshape(-1, D.shape[1], X.shape[1])
    return lab.einsum("ij, il, ik->ljk", D, data_mvp(w, X, D) - y, X).reshape(*v.shape)


def hessian_mvp(v: lab.Tensor, X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    w = v.reshape(-1, D.shape[1], X.shape[1])
    return lab.einsum("ij, nkj, ik, il, im ->nlm", X, w, D, D, X).reshape(*v.shape)


def bd_hessian_mvp(v: lab.Tensor, X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    w = v.reshape(-1, D.shape[1], X.shape[1])
    return lab.einsum("ij, nkj, ik, il->nkl", X, w, D, X).reshape(*v.shape)


# builders


def data(X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return np.concatenate(lab.einsum("ij, ik->jik", D, X), axis=1)


def hessian(X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return lab.einsum("ij, ik, il, im -> jkml", D, D, X, X)


def bd_hessian(X: lab.Tensor, D: lab.Tensor) -> lab.Tensor:
    return lab.einsum("ij, il, im -> jml", D, X, X)



"""Convex re-formulations of neural network models."""

from .l2_regression import LinearRegression
from .logistic_regression import LogisticRegression

__all__ = [
    "LinearRegression",
    "LogisticRegression",
]



"""CVXPY solvers for the cone decomposition problem."""

from typing import Dict, Any, Tuple

from logging import root, INFO
import numpy as np
import cvxpy as cp

import lab

from scnn.private.models import Model, ConvexMLP, AL_MLP

from .cvxpy_solver import CVXPYSolver


class DecompositionProgram(CVXPYSolver):
    """Convert a Gated ReLU model into a ReLU model by cone decomposition.

    Base class for CVXPY programs that convert a convex Gated ReLU model into a
    convex ReLU model by decomposing the weights onto the constraint cones
    :math:`K_i` and :math:`-K_i`.
    """

    def get_objective(self, V, U) -> cp.Minimize:
        raise NotImplementedError(
            "All decomposition programs must implement an objective."
        )

    def __call__(
        self, model: Model, X: np.ndarray, y: np.ndarray
    ) -> Tuple[Model, Dict[str, Any]]:
        """Call the CVXPY solver to compute the decomposition."""

        assert isinstance(model, ConvexMLP)

        # lookup problem dimensions
        n, d = X.shape
        c = y.shape[1]
        p = model.D.shape[1]

        # Cast variables to NumPy arrays.
        X_np = lab.to_np(X)
        D_np = lab.to_np(model.D)
        W = lab.to_np(model.weights.reshape(p * c, d))

        # create optimization variables
        V = cp.Variable((p * c, d))
        U = W + V

        # minimize two-norm of decompositions
        objective = self.get_objective(V, U)

        # constraints
        A = 2 * D_np - np.ones_like(D_np)

        constraints = [
            cp.multiply(A, X_np @ U[i * p : (i + 1) * p].T) >= 0
            for i in range(c)
        ]

        constraints += [
            cp.multiply(A, X_np @ V[i * p : (i + 1) * p].T) >= 0
            for i in range(c)
        ]

        problem = cp.Problem(objective, constraints)

        verbose = root.level <= INFO
        # solve the optimization problem
        problem.solve(
            solver=self.solver, verbose=verbose, **self.solver_kwargs
        )

        # extract solution
        decomp_weights = lab.stack(
            [
                lab.tensor(U.value, dtype=lab.get_dtype()).reshape(c, p, d),
                lab.tensor(V.value, dtype=lab.get_dtype()).reshape(c, p, d),
            ]
        )
        relu_model = AL_MLP(
            d,
            model.D,
            model.U,
            model.kernel,
            regularizer=model.regularizer,
            c=model.c,
        )
        relu_model.weights = decomp_weights

        # extract solver information
        exit_status = {
            "success": problem.status == cp.OPTIMAL,
            "status": problem.status,
            "final_val": problem.value,
        }

        return relu_model, exit_status


class MinL2Decomposition(DecompositionProgram):
    """Decompose by minimize the sum of L2 norms."""

    def get_objective(self, V, U) -> cp.Minimize:
        return cp.Minimize(
            cp.sum(cp.pnorm(U, p=2, axis=1) + cp.pnorm(V, p=2, axis=1))
        )


class MinRelaxedL2Decomposition(DecompositionProgram):
    """Decompose by minimize the sum of L2 norms."""

    def get_objective(self, V, U) -> cp.Minimize:
        return cp.Minimize(cp.sum(V ** 2))


class MinL1Decomposition(DecompositionProgram):
    """Decompose by minimize the sum of L2 norms."""

    def get_objective(self, V, U) -> cp.Minimize:
        return cp.Minimize(
            cp.sum(cp.pnorm(U, p=1, axis=1) + cp.pnorm(V, p=1, axis=1))
        )


class FeasibleDecomposition(DecompositionProgram):
    """Decompose by minimize the sum of L2 norms."""

    def get_objective(self, V, U) -> cp.Minimize:
        return cp.Minimize(1)


class SOCPDecomposition(CVXPYSolver):
    """Base class for CVXPY programs that convert a convex gatedReLU model into
    a convex ReLU model by decomposing the weights onto the constraint cones
    :math:`K_i` and :math:`-K_i`."""

    def __call__(
        self, model: Model, X: np.ndarray, y: np.ndarray
    ) -> Tuple[Model, Dict[str, Any]]:
        """Call the CVXPY solver to compute the decomposition."""

        assert isinstance(model, ConvexMLP)
        if isinstance(model, (AL_MLP)):
            raise ValueError(
                "Cone decomposition is only defined for instances of 'ConvexMLP'"
            )

        # lookup problem dimensions
        n, d = X.shape
        c = y.shape[1]
        p = model.D.shape[1]

        # Cast variables to NumPy arrays.
        X_np = lab.to_np(X)
        D_np = lab.to_np(model.D)
        W = lab.to_np(model.weights.reshape(p * c, d))

        # create optimization variables
        V = cp.Variable((p * c, d))
        U = W + V

        # linear components of the SOCP
        t = cp.Variable((p * c))
        l = cp.Variable((p * c))

        # linear objective
        objective = cp.Minimize(cp.sum(t + l))

        # SOCP constraints
        constraints = [cp.SOC(t, V, axis=1)]
        constraints += [cp.SOC(l, U, axis=1)]

        # cone constraints
        A = 2 * D_np - np.ones_like(D_np)

        constraints += [
            cp.multiply(A, X_np @ U[i * p : (i + 1) * p].T) >= 0
            for i in range(c)
        ]

        constraints += [
            cp.multiply(A, X_np @ V[i * p : (i + 1) * p].T) >= 0
            for i in range(c)
        ]

        problem = cp.Problem(objective, constraints)

        verbose = root.level <= INFO
        # solve the optimization problem
        problem.solve(
            solver=self.solver, verbose=verbose, **self.solver_kwargs
        )

        # extract solution
        decomp_weights = lab.stack(
            [
                lab.tensor(U.value, dtype=lab.get_dtype()).reshape(c, p, d),
                lab.tensor(V.value, dtype=lab.get_dtype()).reshape(c, p, d),
            ]
        )
        relu_model = AL_MLP(
            d,
            model.D,
            model.U,
            model.kernel,
            regularizer=model.regularizer,
            c=model.c,
        )
        relu_model.weights = decomp_weights

        # extract solver information
        exit_status = {
            "success": problem.status == cp.OPTIMAL,
            "status": problem.status,
            "final_val": problem.value,
        }

        return relu_model, exit_status



"""Interfaces for external optimization routines / solvers.

TODO:
    - Extend LinearSolver to support vector-target problems by simultaneously solving each regression problem.
"""

from typing import Dict, Any, Tuple, Optional

import numpy as np

from scnn.private.models import Model, ConvexMLP
from scnn.private.utils.linear import iterative_solvers, preconditioners


class ExternalSolver:
    """Base class / interface for solvers that call external optimization
    routines (e.g. scipy.optimize).

    Subclasses should use '__call__' as an interface to the external routine.
    """

    def __call__(
        self, model: Model, X: np.ndarray, y: np.ndarray
    ) -> Tuple[Model, Dict[str, Any]]:
        """Call the external solver to fit a model."""

        raise NotImplementedError(
            "'ExternalSolver' subclasses must implement '__call__' as an interface to the external solver."
        )


class LinearSolver(ExternalSolver):

    """Interface to scipy.sparse.linalg' solvers for least-squares problems."""

    def __init__(
        self,
        method_name: str,
        max_iters: int,
        tol: float,
        preconditioner: Optional[str] = None,
    ):

        self.method_name = method_name
        self.max_iters = max_iters
        self.tol = tol
        self.preconditioner = preconditioner

    def __call__(
        self,
        model: Any,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[Model, Dict[str, Any]]:

        y = y.squeeze()
        if len(y.shape) > 1:
            raise ValueError(
                "LinearSolver only supports scalar output problems at the moment."
            )

        lam: float = 0.0
        if model.regularizer is not None:
            lam = model.regularizer.lam * len(y)

        M = None
        D = None
        if isinstance(model, ConvexMLP):
            D = model.D

        if self.preconditioner is not None:
            M = preconditioners.get_preconditioner(self.preconditioner, X, D)

        X_op = model.data_operator(X)

        optimal_weights, exit_status = iterative_solvers.lstsq_iterative_solve(
            X_op, y, lam, M, self.method_name, self.max_iters, self.tol
        )

        # reshape weights if necessary
        model.weights = optimal_weights.reshape(model.weights.shape)

        return model, exit_status



"""Functions for generating synthetic datasets.

Public Functions:

    gen_classification_data: generate a synthetic classification dataset with
        targets given by a random neural network.

    gen_regression_data: generate a synthetic regression dataset with targets
        given by a noisy linear model.

    gen_sparse_regression_problem: generate synthetic data from a variety of
        simple feature-sparse planted models.
"""

from typing import Tuple, Literal, Optional, Union, Callable
import math

import numpy as np
from scipy.stats import ortho_group  # type: ignore

Transform = Literal["cosine", "polynomial"]

# local types

Dataset = Tuple[np.ndarray, np.ndarray]


def gen_classification_data(
    data_seed: int,
    n: int,
    n_test: int,
    d: int,
    hidden_units: int = 50,
    kappa: float = 1.0,
) -> Tuple[Dataset, Dataset]:
    """Create a binary classification dataset with a random Gaussian design
    matrix and targets given by a two-layer neural network with random Gaussian
    weights.

    If `kappa` is supplied, then the design matrix satisfies
    :math:`\\kappa(X) \\approx \\text{kappa}`.

    Args:
        data_seed: the seed to use when generating the synthetic dataset.
        n: number of examples in dataset.
        n_test: number of test examples.
        d: number of features for each example.
        hidden_units: (optional) the number of hidden units in the neural
            network.
        kappa: (optional) the (approximate) condition number of the train/test
            design matrices.

    Returns:
        A training set `(X_train, y_train)` and test set `(X_test, y_test)`.
    """
    rng = np.random.default_rng(seed=data_seed)

    w = rng.random(d * hidden_units + hidden_units)
    w1 = w[: d * hidden_units].reshape(hidden_units, d)
    w2 = w[d * hidden_units :].reshape(1, hidden_units)

    Sigma = sample_covariance_matrix(rng, d, kappa)

    X = []
    y = []
    n_pos, n_neg = 0, 0
    n_total = n + n_test

    # simple rejection sampling
    while n_pos + n_neg < n_total:
        xi = rng.multivariate_normal(np.zeros(d), cov=Sigma)
        # compute forward pass
        yi = np.maximum(xi @ w1.T, 0) @ w2.T
        if yi <= 0 and n_neg < math.ceil(n_total):
            y.append(-1)
            X.append(xi)
            n_neg += 1
        elif yi > 0 and n_pos < math.ceil(n_total):
            y.append(1)
            X.append(xi)
            n_pos += 1

    X_np = np.array(X)
    y_np = np.array(y).reshape(-1, 1)

    # shuffle dataset.
    indices = np.arange(n_total)
    rng.shuffle(indices)
    X_np, y_np = X_np[indices], y_np[indices]

    train_set = (X_np[:n], y_np[:n])
    test_set = (X_np[n:], y_np[n:])

    return train_set, test_set


def gen_regression_data(
    data_seed: int,
    n: int,
    n_test: int,
    d: int,
    c: int = 1,
    sigma: float = 0,
    kappa: float = 1,
) -> Tuple[Dataset, Dataset, np.ndarray]:
    """Create a regression dataset with a random Gaussian design matrix.

    If `kappa` is supplied, then the design matrix satisfies
    :math:`\\kappa(X) \\approx \\text{kappa}`.

    Args:
        data_seed: the seed to use when generating the synthetic dataset.
        n: number of examples in dataset.
        n_test: number of test examples.
        d: number of features for each example.
        c: dimension of the targets.
            Defaults to scalar targets (`c = 1`).
        sigma: variance of (Gaussian) noise added to targets.
            Defaults to `0` for a noiseless model.
        kappa: condition number of :math:`\\mathbb{E_X}[X^\\top X]`.
            Defaults to 1.

    Returns:
        A training set `(X_train, y_train)`, test set `(X_test, y_test)`, and
        the group-truth model `w_opt`.
    """
    rng = np.random.default_rng(seed=data_seed)
    # sample "true" model
    w_opt = rng.standard_normal((d, c))
    Sigma = sample_covariance_matrix(rng, d, kappa)

    X = rng.multivariate_normal(np.zeros(d), cov=Sigma, size=n + n_test)
    y = np.dot(X, w_opt)

    if sigma != 0:
        y = y + rng.normal(0, scale=sigma)

    train_set = (X[:n], y[:n])
    test_set = (X[n:], y[n:])

    return train_set, test_set, w_opt


def gen_sparse_regression_problem(
    data_seed: int,
    n: int,
    n_test: int,
    d: int,
    sigma: float = 0,
    kappa: float = 1,
    nnz: Optional[int] = None,
    transform: Optional[Union[Transform, Callable]] = None,
) -> Tuple[Dataset, Dataset, np.ndarray]:
    """Sample data from a feature-sparse planted model.

    Create a realizable regression problem by sampling data from a
    simple planted model. A variety of planted models are available; see
    the `transform` argument. Inspired by code form Tolga Ergen.

    Args:
        data_seed: the seed to use when generating the synthetic dataset.
        n: number of examples in dataset.
        n_test: number of test examples.
        d: number of features for each example.
        sigma: variance of (Gaussian) noise added to targets.
            Defaults to `0` for a noiseless model.
        kappa: condition number of E[X.T X].
            Defaults to 1.
        nnz: number of non-zeros zeros in the true model.
        transform: a non-linear transformation. This must be `None`,
            `'cosine'`, `'polynomial'`, `'product'`, or a callable function that applies a
            custom transformation.

    Returns:
        A training set `(X_train, y_train)`, test set `(X_test, y_test)`, and
        the group-truth model `w_opt`.
    """

    rng = np.random.default_rng(seed=data_seed)
    # sample "true" model
    w_opt = rng.standard_normal((d))

    # true model is sparse
    non_zero_indices = np.arange(d)
    if nnz is not None:
        non_zero_indices = rng.choice(d, size=nnz, replace=False)
        mask = np.zeros_like(w_opt)
        mask[non_zero_indices] = 1.0
        w_opt = w_opt * mask

    # sample covariance matrix.
    Sigma = sample_covariance_matrix(rng, d, kappa)

    X = rng.multivariate_normal(np.zeros(d), cov=Sigma, size=n + n_test)
    y = np.dot(X, w_opt)

    if transform is None:
        # no transform
        y = y
    elif transform == "cosine":
        # cosine transform
        y = np.cos(y)
    elif transform == "cubic":
        # simple cubic
        y = y + (y ** 2) / 2 + (y ** 3) / 6
    elif transform == "product":
        y = np.sign(np.prod(X[:, non_zero_indices], axis=1))
    else:
        try:
            y = transform(y)
        except TypeError:
            raise ValueError(
                f"Transform {transform} not recognized. It must be a \
                        predefined transform, a callable function, or `None`."
            )

    # add noise
    if sigma != 0:
        y = y + rng.normal(0, scale=sigma)

    train_set = (X[:n], y[:n])
    test_set = (X[n:], y[n:])

    return train_set, test_set, w_opt


def sample_covariance_matrix(
    rng: np.random.Generator, d: int, kappa: float
) -> np.ndarray:
    """Sample a covariance matrix with a specific condition number.

    This functions samples a symmetric positive-definite matrix
    with condition number exactly `kappa`. The minimum eigenvalue is `1` and
    the maximum is `kappa`, while the remaining eigenvalues are distributed
    uniformly at random in the interval `[1, kappa]`.

    Args:
        rng: a NumPy random number generator.
        d: the dimensionality of the covariance matrix.
        kappa: condition number of the covariance matrix.

    Returns:
        A (n,d) matrix :math:`\\Sigma` with condition number
            `kappa`.
    """

    # sample random orthonormal matrix
    Q = ortho_group.rvs(d, random_state=rng)
    # sample eigenvalues so that lambda_1 / lambda_d = kappa.
    eigs = rng.uniform(low=1, high=kappa, size=d - 2)
    eigs = np.concatenate([np.array([kappa, 1]), eigs])
    # compute covariance
    Sigma = np.dot(Q.T, np.multiply(np.expand_dims(eigs, axis=1), Q))

    return Sigma



"""Functions for interfacing between the public API and the models, optimizers,
and data representations used in the :module:`private`."""

from .data import (
    normalized_into_input_space,
    input_into_normalized_space,
    process_data,
)

from .models import (
    build_internal_regularizer,
    build_internal_model,
    update_public_model,
    build_public_model,
)

from .solvers import (
    build_prox_operator,
    build_fista,
    build_optimizer,
)

from .metrics import (
    build_metrics_tuple,
    update_public_metrics,
)

from .utils import (
    get_logger,
    set_device,
)

__all__ = [
    "normalized_into_input_space",
    "input_into_normalized_space",
    "process_data",
    "build_internal_regularizer",
    "build_internal_model",
    "update_public_model",
    "build_public_model",
    "build_prox_operator",
    "build_fista",
    "build_optimizer",
    "build_metrics_tuple",
    "update_public_metrics",
    "get_logger",
    "set_device",
]



"""Optimization procedure for PyTorch models."""
from logging import Logger, root, INFO
from typing import Dict, Any, List, Tuple, Callable, Optional
import timeit

from tqdm.auto import tqdm  # type: ignore
import torch
import numpy as np

import lab

from scnn.private.methods.optimization_procedures.optimization_procedure import (
    METRIC_FREQ,
    EPOCH_LOG_FREQ,
)
from scnn.private.methods.optimization_procedures.iterative import (
    IterativeOptimizationProcedure,
)

from scnn.private.methods.optimizers import Optimizer
from scnn.private.models.model import Model
from scnn.private.metrics import update_metrics, init_metrics, format_recent_metrics
from scnn.private.methods.termination_criteria import TerminationCriterion
import scnn.private.loss_functions as loss_fns


class TorchLoop(IterativeOptimizationProcedure):

    """Optimization loop for PyTorch optimizers and models."""

    optimizer: torch.optim.Optimizer

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        max_epochs: int,
        batch_size: int,
        term_criterion: TerminationCriterion,
        name: str = "",
        divergence_check: bool = True,
        log_freq: int = EPOCH_LOG_FREQ,
        metric_freq: int = METRIC_FREQ,
        scheduler: Optional[Callable] = None,
        pre_process: Optional[Callable] = None,
        post_process: Optional[Callable] = None,
    ):
        """
        :param optimizer: an optimizer instance implementing the 'step' method.
        :param max_epochs: the maximum number of epochs to run the optimizer.
        :param batch_size: the batch_size to use.
        :param term_criterion: the criterion to use when checking for early termination.
        :param name: an optional name to display when printing the progress bar.
        :param divergence_check: an whether or not the optimization procedure should
            check for divergence behavior and terminate early.
        :param log_freq: the frequency at which to information during optimization.
        :param metric_freq: the frequency (in epochs) at which to collect metrics, like objective,
            gradient-norm, etc, during optimization.
        :param scheduler: a PyTorch-style step-size scheduler.
        :param pre_process: (optional) a function to call on the model after the it is
            initialized but *before* optimization starts.
        :param post_process: (optional) a function to call on the model *after* optimization
            is complete.
        """
        self.optimizer = optimizer
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.term_criterion = term_criterion

        self.name = name
        self.divergence_check = divergence_check

        self.log_freq = log_freq
        self.metric_freq = metric_freq
        self.scheduler = scheduler

        self.pre_process = pre_process
        self.post_process = post_process

    def reset(self):
        """Reset the optimization procedure (including any attached optimizers)
        to their original state."""
        # TODO: reset the optimizer.

    def _get_optimizer_step_size(self):
        return self.optimizer.state_dict()["param_groups"][0]["lr"]

    def __call__(
        self,
        logger: Logger,
        model: Model,
        initializer: Callable[[Model], Model],
        train_data: Tuple[torch.Tensor, torch.Tensor],
        test_data: Tuple[torch.Tensor, torch.Tensor],
        metrics: Tuple[List[str], List[str], List[str]],
        final_metrics: Optional[Tuple[List[str], List[str], List[str]]] = None,
        callback: Callable = None,
    ) -> Tuple[Dict[str, Any], Model, Dict[str, Any]]:
        """Optimize a model using an iterative optimize method by running a
        "optimization" loop. The loop runs for 'max_iters' number of
        iterations, collects optional metrics at every iteration, and will
        terminate early if the conditions of 'term_criterion' are met.

        :param logger: a logging.Logger object that can be used to log information during
            optimization.
        :param model: the model to optimize.
        :param initializer: a function that can be called to initialize the model.
        :param train_data: an (X, y) tuple representing the training set.
        :param test_data: an (X_test, y_test) tuple representing the test data.
        :param metrics: a tuple of the form (train_metrics, test_metrics, additional_metrics)
            specifying the metrics to be computed on the training set, test set, and data-independent
            metrics.
        :returns: (exit_status, model, metrics) --- execution information, the optimized model, and metrics from optimization.
        """
        # enable autodiff engine
        lab.toggle_autodiff(True)

        # pre-optimization setup
        with torch.no_grad():
            (
                model,
                metrics_log,
                exit_status,
                objective,
                grad,
            ) = self._pre_optimization(
                logger,
                model,
                initializer,
                train_data,
                test_data,
                metrics,
                final_metrics,
                self._get_optimizer_step_size(),
            )
            (X, y), f0, start_time = train_data, objective, None

        def seed_worker(worker_id):
            worker_seed = torch.initial_seed() % 2 ** 32
            np.random.seed(worker_seed)

        tensor_dataset = torch.utils.data.TensorDataset(X, y)
        training_loader = torch.utils.data.DataLoader(
            tensor_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            worker_init_fn=seed_worker,
            generator=lab.torch_rng,
        )

        # training loop
        verbose = root.level <= INFO
        for epoch in tqdm(
            range(self.max_epochs), desc=self.name, disable=(not verbose)
        ):
            start_time = self._get_start_time(start_time)

            if epoch % self.log_freq == 0 and verbose:
                tqdm.write(format_recent_metrics(metrics_log, metrics))

            # check termination criteria
            with torch.no_grad():
                if self.term_criterion(model, X, y, objective, grad):
                    exit_status["success"] = True
                    logger.info(
                        f"Termination criterion satisfied at iteration {epoch}/{self.max_epochs}. Exiting optimization loop."
                    )
                    break

                # check for divergence
                if (
                    self.divergence_check
                    and objective > 1000 * f0
                    or torch.isnan(objective)
                ):
                    exit_status["success"] = False
                    logger.warning(
                        f"Method diverged at {epoch}/{self.max_epochs}. Exiting optimization loop."
                    )
                    break

            for itr, (xi, yi) in enumerate(training_loader):

                self.optimizer.zero_grad()
                # hard-code squared error at the moment.
                obj = model.objective(xi, yi)
                obj.backward()
                self.optimizer.step()

            if callback is not None:
                model = callback(model, X, y)

            # call the scheduler once per epoch in the PyTorch style.
            if self.scheduler is not None:
                self.scheduler.step()

            # update metrics every epoch.
            if epoch % self.metric_freq == 0:
                with torch.no_grad():

                    # calculate time for one iteration.
                    metrics_log, start_time = self._record_time(metrics_log, start_time)
                    objective = model.objective(X, y)
                    grad = model.grad(X, y)
                    metrics_log = update_metrics(
                        metrics_log,
                        model,
                        {"step_size": self._get_optimizer_step_size()},
                        train_data,
                        test_data,
                        metrics,
                        objective,
                        grad,
                    )

        # post-optimization clean-up.
        model, exit_status, metrics_log = self._post_optimization(
            logger,
            model,
            train_data,
            test_data,
            metrics,
            final_metrics,
            exit_status,
            metrics_log,
            start_time,
            self._get_optimizer_step_size(),
        )

        # collect run-information
        if epoch == self.max_epochs - 1:
            exit_status["success"] = False
            logger.warning(
                "Max iterations reached before termination criterion was satisfied. Exiting optimization loop."
            )
        exit_status["epochs"] = epoch + 1

        # disable autodiff engine
        lab.toggle_autodiff(False)

        return exit_status, model, metrics_log



"""Implementation of one-layer Gated ReLU MLP with squared-error objective."""

from typing import Optional

import lab

from .relu_mlp import ReLUMLP
from scnn.private.models.regularizers import Regularizer
from scnn.private.loss_functions import squared_error, relu


class GatedReLUMLP(ReLUMLP):

    """One-layer Gated ReLU MLP with squared-error objective."""

    def __init__(
        self,
        d: int,
        U: lab.Tensor,
        regularizer: Optional[Regularizer] = None,
        c: int = 1,
    ):
        """
        :param d: the dimensionality of the dataset (ie.number of features).
        :param U: the gate vectors associated with gated ReLU activations.
        :param regularizer: (optional) a penalty function controlling the flexibility of the model.
        :param c: (optional) the number of targets.
        """

        super().__init__(d, U.shape[1], regularizer, c=c)
        self.U = U

    def _forward(self, X: lab.Tensor, w: lab.Tensor, **kwargs) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the forward pass.
        """
        w1, w2 = self._split_weights(w)

        return lab.multiply(lab.sign(relu(X @ self.U)), X @ w1.T) @ w2.T

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: parameter at which to compute the objective.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: objective L(f, (X, y)).
        """
        return squared_error(self._forward(X, w), y) / (
            2 * self._scaling(y, scaling)
        )

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the l2 objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: parameter at which to compute the gradient pass.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        w1, w2 = self._split_weights(w)
        D = lab.sign(relu(X @ self.U))
        Z = lab.multiply(D, X @ w1.T)
        residuals = Z @ w2.T - y

        return self._grad_helper(X, residuals, w1, w2, D, Z) / self._scaling(
            y, scaling
        )

    def _grad_helper(
        self,
        X: lab.Tensor,
        residuals: lab.Tensor,
        w1: lab.Tensor,
        w2: lab.Tensor,
        D: lab.Tensor,
        Z: lab.Tensor,
    ) -> lab.Tensor:

        # always uses the 0 subgradient of max(x, 0) when x = 0.
        g1 = lab.einsum("ij, il, ik, lj->jk", D, residuals, X, w2)
        g2 = Z.T @ residuals

        return self._join_weights(g1, g2.T)



"""Data transformations and related utilities.

Public Functions:

    unitize_columns: unitize the columns of a training set, optionally, a test set.


TODO:
    - Cleanup comments in this file.
"""

from typing import Tuple, Optional

import lab
import numpy as np


Dataset = Tuple[lab.Tensor, lab.Tensor]


def add_bias_col(dataset: Dataset) -> Dataset:
    """Augment the feature matrix in a dataset with a bias column.

    Args:
        dataset: a (X, y) tuple defining a dataset.
    """

    X, y = dataset

    X = lab.concatenate([X, lab.ones((X.shape[0], 1))], axis=1)

    return (X, y)


def unitize_columns(
    train_set: Dataset,
    test_set: Optional[Dataset] = None,
) -> Tuple[Dataset, Dataset, lab.Tensor]:
    """Transform a dataset so that the columns of the design matrix have unit
    norm,

    .. math::

        \\text{diag} (\\tilde X^\\top \\tilde X) = I

    If a test set is also provided, the column-norms of the training set are used to apply the same transformation to the test data.

    Args:
        train_set: an (X, y) tuple.
        test_set: (optional) an (X_test, y_test) tuple.

    Returns:
       (train_set, test_set, column_norms) --- a tuple containing the transformed training set, test set, and the column norms of the training design matrix.
       If a test_set is not provided, then `test_set` is None.
    """

    X_train = train_set[0]
    column_norms = lab.sqrt(lab.sum(X_train ** 2, axis=0, keepdims=True))

    X_train = X_train / column_norms
    train_set = (X_train, train_set[1])

    if test_set is not None:
        test_set = (test_set[0] / column_norms, test_set[1])

    return train_set, test_set, column_norms


def train_test_split(X, y, valid_prop=0.2, split_seed=1995):
    """ """
    n = y.shape[0]
    split_rng = np.random.default_rng(seed=split_seed)
    num_test = int(np.floor(n * valid_prop))
    indices = np.arange(n)
    split_rng.shuffle(indices)
    test_indices = indices[:num_test].tolist()
    train_indices = indices[num_test:].tolist()

    # subset the dataset
    X_train = X[train_indices, :]
    y_train = y[train_indices]

    X_test = X[test_indices, :]
    y_test = y[test_indices]

    return (X_train, y_train), (X_test, y_test)



"""Wrapper for PyTorch layers allowing for (optional) layer-wise
regularizers."""
from typing import Optional

import torch

from scnn.private.models.regularizers import Regularizer


class LayerWrapper(torch.nn.Module):
    """Wrapper for Pytorch module which provides the opportunity for built-in
    regularization.

    Useful in the case of layer-wise regularization rather than over the whole
    network
    """

    def __init__(self, layer, regularizer: Optional[Regularizer] = None):

        super().__init__()
        self.layer = layer
        self.regularizer = regularizer

    def forward(self, X: torch.Tensor):
        return self.layer.forward(X)

    def get_regularizer(self):
        return self.regularizer

    def parameters(self):
        return torch.cat([param for param in self.layer.parameters()])



"""Preconditioners for iterative linear-system solvers."""

from typing import Optional, Union

from scnn.private.utils.linear_operators import MatVecOperator
import lab


def get_preconditioner(
    name: str, X: lab.Tensor, D: Optional[lab.Tensor] = None
) -> Union[MatVecOperator, lab.Tensor]:
    """Lookup and return preconditioner.

    :param name: the name of the preconditioner to return.
    :param X: (n,d) feature matrix of training examples.
    :param D: (optional) a matrix of sign patterns for the expanded
    feature matrix.
    """
    if name == "column_norm":
        return column_norm(X, D)
    elif name == "H_diag":
        n = X.shape[0]
        return hessian_diagonal(X, D) / n
    else:
        raise ValueError(f"Preconditioner {name} not recognized!")


def hessian_diagonal(X: lab.Tensor, D: Optional[lab.Tensor] = None) -> lab.Tensor:
    """Compute the diagonal of the Hessian for the squared loss. If D is not
    None, then the diagonal the expanded matrix.

    .. math::

        [diag(D[:,1] X, ...., diag(D[:,P]) X],

    is computed.
    :param X: the (n,d) matrix of data examples.
    :param D: (optional) a (d, p) associated with a convex neural network. See 'models/nns.py'.
    :returns: diagonal of the Hessian matrix as a vector.
    """

    if D is not None:
        # compute column norms of expanded matrix.
        H_diag = lab.einsum("ij, ik->jk", D, X ** 2)
    else:
        H_diag = lab.sum(X ** 2, axis=0)

    return H_diag


def column_norm(X: lab.Tensor, D: Optional[lab.Tensor] = None) -> MatVecOperator:
    """Compute a diagonal preconditioner that unitizes the columns of 'X', as described
    here: https://web.stanford.edu/group/SOL/software/lsmr/.
    If D is not None, then a preconditioner for the expanded matrix
    .. math::

        [diag(D[:,1] X, ...., diag(D[:,P]) X],

    is computed.
    :param X: the (n,d) matrix of data examples.
    :param D: (optional) a (d, p) associated with a convex neural network. See 'models/nns.py'.
    :returns: (forward_map, inverse_map, column_norms)
    """

    H_diag = hessian_diagonal(X, D).reshape(-1)
    col_norms = lab.sqrt(H_diag)

    # create operators for forward and inverse maps.
    def forward_map(v: lab.Tensor):
        v = v.reshape(-1)
        return lab.safe_divide(v, col_norms)

    s = col_norms.shape[0]

    # both operators are symmetric and real, so the Hermitian/transpose is the same operator.
    forward = MatVecOperator(shape=(s, s), forward=forward_map, transpose=forward_map)

    return forward



"""Implementations of non-convex models."""

from .torch import SequentialWrapper, LayerWrapper, GatedReLULayer
from .manual import ReLUMLP, GatedReLUMLP

__all__ = [
    "SequentialWrapper",
    "LayerWrapper",
    "GatedReLULayer",
    "ReLUMLP",
    "GatedReLUMLP",
]



"""Interface for model classes."""

from typing import Optional, Union

import lab


class Regularizer:

    """Base class for regularizers."""

    lam: float

    def __init__(
        self,
        lam: float,
    ):
        """
        :param base_model: the Model instance to regularize.
        :param lam: the tuning parameter controlling the strength of regularization.
        """

        self.lam = lam

    # abstract methods that must be overridden

    def penalty(self, w: lab.Tensor, **kwargs) -> float:
        """Compute the penalty associated with the regularizer.

        :param w: parameter at which to compute the penalty value.
        :returns: penalty value
        """

        raise NotImplementedError("A regularizer must implement 'penalty'!")

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        **kargs,
    ) -> lab.Tensor:
        """Compute the gradient of the regularizer.

        :param w: parameter at which to compute the penalty gradient.
        :param base_grad: (optional) the gradient of the un-regularized objective. This is
            used to compute the minimum-norm subgradient for "pseudo-gradient" methods.
        :returns: gradient.
        """

        raise NotImplementedError("A regularizer must implement 'grad'!")



"""Regularizers."""

from .regularizer import Regularizer
from .constraint import Constraint
from .group_l1 import GroupL1Regularizer
from .group_l1_orthant import GroupL1Orthant
from .feature_group_l1 import FeatureGroupL1Regularizer
from .l2 import L2Regularizer
from .l1 import L1Regularizer
from .orthant import OrthantConstraint
from .l1_squared import L1SquaredRegularizer

__all__ = [
    "Regularizer",
    "Constraint",
    "GroupL1Regularizer",
    "GroupL1Orthant",
    "FeatureGroupL1Regularizer",
    "L2Regularizer",
    "L1Regularizer",
    "OrthantConstraint",
    "L1SquaredRegularizer",
]



"""Gradient descent with line-search."""
from typing import Callable, Tuple, Dict, Any, Optional

import lab

from scnn.private.methods.core.line_search import ls
from scnn.private.methods.line_search import Backtracker, LSCondition
from scnn.private.prox import ProximalOperator


def gradient_step(
    w: lab.Tensor,
    grad: lab.Tensor,
    step_size: float,
    prox: Optional[ProximalOperator] = None,
) -> lab.Tensor:
    """Take one step of gradient descent.

    :param w: the parameters to be updated.
    :param grad: the gradient of the objective function.
    :param step_size: the step-size to use.
    :param prox: NOT USED. Included to unify step signatures.
    :returns: updated parameters
    """

    return w - step_size * grad


def gd_ls(
    w: lab.Tensor,
    f0: float,
    descent_dir: lab.Tensor,
    grad: lab.Tensor,
    obj_fn: Callable,
    grad_fn: Callable,
    init_step_size: float,
    ls_cond: LSCondition,
    backtrack: Backtracker,
    prox: Optional[ProximalOperator] = None,
) -> Tuple[lab.Tensor, float, float, Dict[str, Any]]:
    """Take one step of gradient descent using a line-search to pick the step-
    size.

    :param w: the parameters to be updated.
    :param f0: the objective function evaluated at w.
    :param descent_dir: the descent direction for the step.
    :param grad: the gradient of the smooth component of the loss function.
    :param obj_fn: function that returns the objective when called at w.
    :param grad_fn: function that returns the gradient of the objective function.
    :param init_step_size: the first step-size to use.
    :param ls_cond: the line-search condition to check.
    :param backtrack: a rule for calculating the next step-size to try.
    :param prox: NOT USED. Included to unify line-search signatures.
    :returns: (w_next, f1, step_size, exit_state): the updated parameters, objective value, new step-size, and exit state.
    """

    return ls(
        gradient_step,
        w,
        f0,
        descent_dir,
        grad,
        obj_fn,
        grad_fn,
        init_step_size,
        ls_cond,
        backtrack,
    )



"""L2 Regularizer."""
from typing import Optional

import lab

from scnn.private.loss_functions import l2_penalty
from scnn.private.models.regularizers.regularizer import Regularizer


class L2Regularizer(Regularizer):

    """L2-regularizer of the form.

    $f(w) = (lambda/2) * ||w||_2^2$
    """

    def penalty(self, w: lab.Tensor, **kwargs) -> float:
        """Compute the penalty associated with the regularizer.

        :param w: parameter at which to compute the penalty.
        :returns: penalty value
        """

        return l2_penalty(w, self.lam)

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the regularizer.

        :param w: parameter at which to compute the penalty gradient.
        :param base_grad: NOT USED. Gradient from the base model.
        :returns: gradient.
        """

        return self.lam * w



"""Model initializers."""
from logging import Logger
from typing import Tuple, Dict, Any, Callable

import numpy as np

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers.l2 import L2Regularizer
from scnn.private.models import ConvexMLP, ReLUMLP
from scnn.activations import sample_gate_vectors, compute_activation_patterns

# constants

# use a random initialization.
RANDOM = "random"
# initialize at 0.
ZERO = "zero"
# gate vectors
GATES = "gates"
# solve the unregularized regression problem using an iterative solver.
LEAST_SQRS = "least_squares"

# default solver config
DEFAULT_LSMR_SOLVER = {"name": "lsmr"}


def get_initializer(
    logger: Logger,
    rng: np.random.Generator,
    train_data: Tuple[lab.Tensor, lab.Tensor],
    config: Dict[str, Any],
) -> Callable[[Model], Model]:
    """Construct and return a closure which can be used to initialize a model
    before optimization.

    :param logger: a logger instance.
    :param rng: a seeded random number generator.
    :param train_data: the training set. This is required for "clever" initializations based on
        solving the un-regularized least-squares problem.
    :param config: the configuration dictionary for the initialization method.
    """

    def initialize_model(model_to_init: Model) -> Model:
        """Initialize the given model according to the given initializer
        config.

        :param model_to_init: the model to initialize.
        :returns: the initialized model.
        """
        name = config.get("name", None)
        lam = 0.0

        if model_to_init.regularizer is not None and hasattr(
            model_to_init.regularizer, "lam"
        ):
            lam = model_to_init.regularizer.lam

        if name is None:
            return model_to_init
        elif name == RANDOM:
            model_to_init.set_weights(
                lab.tensor(
                    rng.standard_normal(model_to_init.weights.shape),
                    dtype=lab.get_dtype(),
                )
            )
        elif name == GATES:
            if hasattr(model_to_init, "U"):
                model_to_init.set_weights(model_to_init.U)
            else:
                act_config = config.get("sign_patterns", {"seed": 778})

                G = sample_gate_vectors(
                    act_config["seed"], train_data[0].shape[1], act_config["n_samples"]
                )
                D, G = compute_activation_patterns(lab.to_np(train_data[0]), G)
                G = lab.tensor(G.T, dtype=lab.get_dtype())

                if isinstance(model_to_init, ReLUMLP):
                    p = G.shape[0]
                    model_to_init.p = p
                    G = lab.concatenate([lab.ravel(G), lab.ones(p)])

                model_to_init.set_weights(G)

        elif name == ZERO:
            if isinstance(model_to_init, ReLUMLP):
                raise ValueError(
                    "A (Gated) ReLU MLPs should never be initialized at zero!"
                )

            model_to_init.set_weights(lab.zeros_like(model_to_init.weights))

        elif name == LEAST_SQRS:
            # TODO: Fixme!
            raise NotImplementedError("TODO! Fix least squares initialization.")

            if not isinstance(model_to_init, ConvexMLP):
                raise ValueError(
                    "Initializing with the ridge-regression solution is only for ConvexMLPs and subclasses."
                )

            cur_backend = lab.backend
            X_np, y_np = lab.all_to_np(train_data)
            y_np = np.squeeze(y_np)
            if len(y_np.shape) > 1:
                raise ValueError(
                    "Cannot use least-squares initialization with multi-target problems!"
                )

            np_train_data = (X_np, y_np)

            np_D = lab.to_np(model_to_init.D)
            np_U = lab.to_np(model_to_init.U)

            # least squares only supports NumPy
            lab.set_backend(lab.NUMPY)
            solver_config = config.get("solver", DEFAULT_LSMR_SOLVER)

            # scale lambda to account for scaling the objective by 1/n.
            scaled_lam = lam * np_train_data[0].shape[0]
            l2 = L2Regularizer(scaled_lam)
            model_to_fit = ConvexMLP(
                model_to_init.d,
                np_D,
                np_U,
                model_to_init.kernel,
                regularizer=l2,
            )

            # solver = get_method(logger, rng, model_to_fit, np_train_data, solver_config)
            solver = None
            exit_status, fit_model, metrics_log = solver(
                logger,
                model_to_fit,
                lambda x: x,
                np_train_data,
                np_train_data,
                ([], [], []),
            )

            # reset the linear algebra backend.
            lab.set_backend(cur_backend)
            fit_model_weights = lab.tensor(fit_model.weights, dtype=lab.get_dtype())

            model_to_init.set_weights(fit_model_weights)

        else:
            raise ValueError(
                f"Initializer {name} not recognized! Please register it in 'models/initializers.py'."
            )

        return model_to_init

    return initialize_model



"""Models."""

from .model import Model
from .linear import LinearRegression, LogisticRegression
from .non_convex import (
    SequentialWrapper,
    LayerWrapper,
    ReLUMLP,
    GatedReLUMLP,
    GatedReLULayer,
)

from .convex import (
    ConvexMLP,
    AL_MLP,
)

from .decompositions import QuadraticDecomposition

from .one_vs_all import OneVsAllModel

from .regularizers import (
    Regularizer,
    Constraint,
    GroupL1Orthant,
    GroupL1Regularizer,
    FeatureGroupL1Regularizer,
    L2Regularizer,
    L1Regularizer,
    OrthantConstraint,
    L1SquaredRegularizer,
)

from .solution_mappings import (
    is_compatible,
    get_nc_formulation,
)


__all__ = [
    "Model",
    "LinearRegression",
    "LogisticRegression",
    "SequentialWrapper",
    "LayerWrapper",
    "ReLUMLP",
    "GatedReLUMLP",
    "GatedReLULayer",
    "ConvexMLP",
    "AL_MLP",
    "QuadraticDecomposition",
    "OneVsAllModel",
    "Regularizer",
    "Constraint",
    "GroupL1Regularizer",
    "FeatureGroupL1Regularizer",
    "GroupL1Orthant",
    "L2Regularizer",
    "L1Regularizer",
    "L1SquaredRegularizer",
    "OrthantConstraint",
    "is_compatible",
    "get_nc_formulation",
]



"""Methods."""

# ===== module exports ===== #

from .optimization_procedures import (
    OptimizationProcedure,
    IterativeOptimizationProcedure,
    TorchLoop,
    DoubleLoopProcedure,
    METRIC_FREQ,
    ITER_LOG_FREQ,
    EPOCH_LOG_FREQ,
)
from .core import (
    ls,
    gd_ls,
    proximal_gradient_step,
    proximal_gradient_ls,
    fista_step,
    fista_ls,
    update_multipliers,
    acc_update_multipliers,
)

from .optimizers import (
    Optimizer,
    ProximalOptimizer,
    ProximalLSOptimizer,
    MetaOptimizer,
    GD,
    GDLS,
    PGD,
    PGDLS,
    FISTA,
    AugmentedLagrangian,
)

from .line_search import (
    Backtracker,
    MultiplicativeBacktracker,
    LSCondition,
    QuadraticBound,
    DiagQB,
    Armijo,
    StepSizeUpdater,
    KeepNew,
    KeepOld,
    ForwardTrack,
    Lassplore,
)

from .callbacks import (
    ObservedSignPatterns,
    ConeDecomposition,
    ProximalCleanup,
    ApproximateConeDecomposition,
)

from .external_solver import LinearSolver

from .cvxpy import (
    CVXPYSolver,
    CVXPYGatedReLUSolver,
    CVXPYReLUSolver,
    MinL2Decomposition,
    MinL1Decomposition,
    FeasibleDecomposition,
    MinRelaxedL2Decomposition,
    SOCPDecomposition,
)

from .termination_criteria import (
    GradientNorm,
    StepLength,
    ConstrainedHeuristic,
    LagrangianGradNorm,
)


__all__ = [
    "OptimizationProcedure",
    "IterativeOptimizationProcedure",
    "TorchLoop",
    "DoubleLoopProcedure",
    "ObservedSignPatterns",
    "ConeDecomposition",
    "ApproximateConeDecomposition",
    "ProximalCleanup",
    "ls",
    "gradient_step",
    "gd_ls",
    "proximal_gradient_step",
    "proximal_gradient_ls",
    "fista_step",
    "fista_ls",
    "update_multipliers",
    "acc_update_multipliers",
    "Optimizer",
    "ProximalOptimizer",
    "ProximalLSOptimizer",
    "MetaOptimizer",
    "GD",
    "GDLS",
    "PGD",
    "PGDLS",
    "FISTA",
    "AugmentedLagrangian",
    "LinearSolver",
    "CVXPYSolver",
    "CVXPYGatedReLUSolver",
    "CVXPYReLUSolver",
    "MinL2Decomposition",
    "MinL1Decomposition",
    "FeasibleDecomposition",
    "MinRelaxedL2Decomposition",
    "SOCPDecomposition",
    "Backtracker",
    "MultiplicativeBacktracker",
    "LSCondition",
    "QuadraticBound",
    "DiagQB",
    "Armijo",
    "StepSizeUpdater",
    "KeepNew",
    "KeepOld",
    "ForwardTrack",
    "Lassplore",
    "GradientNorm",
    "StepLength",
    "ConstrainedHeuristic",
    "LagrangianGradNorm",
]



"""Utilities for interfacing between public API and
:module:`scnn.private`."""

import logging
import lab


def get_logger(
    name: str, verbose: bool = False, debug: bool = False, log_file: str = None
) -> logging.Logger:
    """Construct a logging.Logger instance with an appropriate configuration.

    Args:
        name: name for the Logger instance.
        verbose: (optional) whether or not the logger should print verbosely (ie. at the INFO level).
            Defaults to False.
        debug: (optional) whether or not the logger should print in debug mode (ie. at the DEBUG level).
            Defaults to False.
        log_file: (optional) path to a file where the log should be stored. The log is printed to stdout when 'None'.

    Returns:
         Instance of logging.Logger.
    """

    level = logging.WARNING
    if debug:
        level = logging.DEBUG
    elif verbose:
        level = logging.INFO

    logging.basicConfig(level=level, filename=log_file)
    logger = logging.getLogger(name)
    logging.root.setLevel(level)
    logger.setLevel(level)
    return logger


def set_device(device: str = "cpu", dtype="float32", seed: int = 778):
    """Set the device to be used by LAB, the global dtype, and random seeds."""
    if device == "cpu":
        lab.set_backend("numpy")
        lab.set_device(device)
    elif device == "cuda":
        lab.set_backend("torch")
        lab.set_device(device)

    lab.set_dtype(dtype)
    lab.set_seeds(seed)
